import dataclasses

import algosdk

from ..transaction_group import TransactionGroup
from ..utils import get_selector, sp_fee
from .base_factory import (
    PoolBuildParams,
    PoolFactory,
    PoolParams,
    get_contract_deploy_cost,
)


def build_constant_product_tx_group(
    factory_id: int,
    sender: str,
    sp: algosdk.transaction.SuggestedParams,
    pool_params: PoolParams,
) -> TransactionGroup:

    deployment_cost = get_contract_deploy_cost(
        is_algo=pool_params.primary_asset_id == 0,
        extra_pages=1,
        num_byte_slices=2,
        num_uint=0,
    )

    fund_tx = algosdk.transaction.PaymentTxn(
        sender=sender,
        receiver=algosdk.logic.get_application_address(factory_id),
        amt=deployment_cost,
        sp=sp,
    )

    app_args: list = [
        get_selector("build(asset,asset,uint64)uint64"),
        algosdk.abi.UintType(8).encode(0),
        algosdk.abi.UintType(8).encode(1),
        algosdk.abi.UintType(64).encode(pool_params.fee_bps),
    ]

    box_name = pool_params.to_box_name()

    build_tx = algosdk.transaction.ApplicationNoOpTxn(
        sender=sender,
        index=factory_id,
        app_args=app_args,
        sp=sp_fee(sp, 10000),
        boxes=[(0, box_name)],
        foreign_assets=[
            pool_params.primary_asset_id,
            pool_params.secondary_asset_id,
        ],
    )

    return TransactionGroup([fund_tx, build_tx])


@dataclasses.dataclass
class ConstantProductFactory(PoolFactory):
    def build_tx_group(
        self,
        sender: str,
        sp: algosdk.transaction.SuggestedParams,
        pool_build_params: PoolBuildParams,
    ):
        return build_constant_product_tx_group(
            factory_id=self.app_id,
            sender=sender,
            sp=sp,
            pool_params=PoolParams(
                **dataclasses.asdict(pool_build_params),
                version=self.state.pool_version,
            ),
        )



from algosdk.v2client.algod import AlgodClient

from ..config import Config
from .escrow import Escrow, fetch_escrow_by_id
from .farm import Farm, fetch_farm_by_id


class PactFarmingClient:
    """An entry point for interacting with the farming SDK."""

    algod: AlgodClient
    """Algorand client to work with."""

    config: Config

    def __init__(self, algod: AlgodClient, config: Config):
        """
        Args:
            algod: Algorand client to work with.
        """
        self.algod = algod
        self.config = config

    def fetch_farm_by_id(self, app_id: int) -> Farm:
        return fetch_farm_by_id(algod=self.algod, app_id=app_id)

    def fetch_escrow_by_id(self, app_id: int) -> Escrow:
        return fetch_escrow_by_id(algod=self.algod, app_id=app_id)



"""This module exposes PactClient class which is an entry points for interacting with the SDK.

Typical usage example::

    import algosdk
    from algosdk.v2client.algod import AlgodClient

    import pactsdk

    algod = AlgodClient("<token>", "<url>")
    pact = pactsdk.PactClient(algod)

    algo = pact.fetch_asset(0)
    other_coin = pact.fetch_asset(12345678)

    pools = pact.fetch_pools_by_assets(algo, other_coin)
"""
from typing import Optional, Union, cast

from algosdk.v2client.algod import AlgodClient

from pactsdk.api import ApiListPoolsResponse
from pactsdk.farming.farming_client import PactFarmingClient

from .asset import Asset, fetch_asset_by_index
from .config import Config, Network, get_config
from .factories import ConstantProductFactory, get_pool_factory
from .gas_station import get_gas_station, set_gas_station
from .pool import (
    ListPoolsParams,
    Pool,
    fetch_pool_by_id,
    fetch_pools_by_assets,
    list_pools,
)


class PactClient:
    """An entry point for interacting with the SDK.

    Exposes convenience methods for fetching assets and pools.
    """

    algod: AlgodClient
    """Algorand client to work with."""

    config: Config
    """Client configuration with global contracts ids etc."""

    farming: PactFarmingClient

    def __init__(self, algod: AlgodClient, network: Network = "mainnet", **kwargs):
        """Constructor for the PactClient class.

        Args:
            algod: Algorand client to work with.
            network: The Algorand network to use the client with. The configuration values depend on the chosen network.
            kwargs: Use it to overwrite configuration parameters.
        """
        self.algod = algod
        self.config = get_config(network, **kwargs)
        self.farming = PactFarmingClient(algod, self.config)

        try:
            get_gas_station()
        except AssertionError:
            set_gas_station(self.config.gas_station_id)

    def fetch_asset(self, asset_index: int) -> Asset:
        """A convenient method for fetching ASAs (Algorand Standard Asset).

        This will return an Asset class with the relevant data about the asset if the asset index is valid.
        Note that an index of zero (0) will return the Algo asset.

        Args:
            asset_index: The id of the asset.

        Raises:
            algosdk.error.AlgodHTTPError: If the asset does not exist.

        Returns:
            Asset instance for the given index.
        """
        return fetch_asset_by_index(self.algod, asset_index)

    def list_pools(
        self, params: Optional[ListPoolsParams] = None
    ) -> ApiListPoolsResponse:
        """Returns a list of pools according to the pool options passed in. Uses Pact API for fetching the data.

        This method is deprecated but is kept for backward compatibility. Pact is in the process of changing the way the pools are created. In the future, all pools will be created using a pool factory contract which allows for an on-chain discoverability of pools.

        Args:
            params: API call parameters.

        Returns:
            Paginated list of pools.
        """
        return list_pools(self.config.api_url, params or {})

    def fetch_pools_by_assets(
        self, primary_asset: Union[Asset, int], secondary_asset: Union[Asset, int]
    ) -> list[Pool]:
        """Returns a list of liquidity pools on Pact that are across the primary and secondary assets.

        First, it uses Pact API retrieve app ids matching the provided assets and then uses algod client to fetch contracts data from the blockchain.

        Args:
            primary_asset: Primary asset or the asset id for the pool to find.
            secondary_asset: Secondary asset or the asset id for the pool to find.

        Returns:
            List of pools for the two assets, the list may be empty.
        """
        return fetch_pools_by_assets(
            algod=self.algod,
            asset_a=primary_asset,
            asset_b=secondary_asset,
            pact_api_url=self.config.api_url,
        )

    def fetch_pool_by_id(self, app_id: int) -> Pool:
        """Fetches the pool by the application id. It uses algod client to fetch the data directly from the blockchain.

        Args:
            app_id: The application id of pool to return.

        Raises:
            algosdk.error.AlgodHTTPError: If the pool does not exist.

        Returns:
            The pool for the application id.
        """
        return fetch_pool_by_id(algod=self.algod, app_id=app_id)

    def get_constant_product_pool_factory(self) -> ConstantProductFactory:
        """Gets the constant product pool factory according to the client's configuration."""
        factory = get_pool_factory(
            algod=self.algod, pool_type="CONSTANT_PRODUCT", config=self.config
        )
        return cast(ConstantProductFactory, factory)

    def get_nft_constant_product_pool_factory(self) -> ConstantProductFactory:
        """Gets the NFT constant product pool factory according to the client's configuration."""
        factory = get_pool_factory(
            algod=self.algod, pool_type="NFT_CONSTANT_PRODUCT", config=self.config
        )
        return cast(ConstantProductFactory, factory)



"""
This module container utilities for interacting with the farm contract.
"""

import datetime
from dataclasses import dataclass, field
from typing import Optional

import algosdk
from algosdk import abi, transaction
from algosdk.v2client.algod import AlgodClient

from pactsdk.asset import Asset, fetch_asset_by_index

from ..encoding import deserialize_uint64
from ..gas_station import get_gas_station
from ..utils import get_selector, parse_app_state, sp_fee
from .escrow import Escrow, build_deploy_escrow_txs, fetch_escrow_by_id
from .farm_state import (
    FarmingRewards,
    FarmInternalState,
    FarmState,
    FarmUserState,
    format_rewards,
    format_rpt,
    internal_state_to_state,
    parse_internal_state,
)

UPDATE_TX_FEE = 3000
MAX_REWARD_ASSETS = 7

UPDATE_GLOBAL_STATE_SIG = get_selector("update_global_state()void")
UPDATE_STATE_SIG = get_selector("update_state(application,account,account,asset)void")
CLAIM_REWARDS_SIG = get_selector("claim_rewards(account,uint64[])void")
ADD_REWARD_ASSET_SIG = get_selector("add_reward_asset(asset)void")
DEPOSIT_REWARDS_SIG = get_selector("deposit_rewards(uint64[],uint64)void")


def fetch_farm_raw_state_by_id(algod: AlgodClient, app_id: int) -> dict:
    app_info = algod.application_info(app_id)
    return parse_app_state(app_info["params"]["global-state"])


def make_farm_from_raw_state(
    algod: AlgodClient, app_id: int, raw_state: dict
) -> "Farm":
    internal_state = parse_internal_state(raw_state)
    state = internal_state_to_state(algod, internal_state)

    return Farm(
        algod=algod,
        app_id=app_id,
        raw_state=raw_state,
        internal_state=internal_state,
        state=state,
    )


def fetch_farm_by_id(algod: AlgodClient, app_id: int):
    raw_state = fetch_farm_raw_state_by_id(algod, app_id)
    return make_farm_from_raw_state(algod, app_id, raw_state)


@dataclass
class Farm:
    algod: AlgodClient
    """The Algorand client to use."""

    app_id: int
    """The application id for the farming."""

    raw_state: dict

    internal_state: FarmInternalState

    state: FarmState

    _suggested_params: Optional[algosdk.transaction.SuggestedParams] = None

    app_address: str = field(init=False)

    def __post_init__(self):
        self.app_address = algosdk.logic.get_application_address(self.app_id)

    def __hash__(self):
        return self.app_id

    def __str__(self):
        return f"Farm {self.staked_asset} ({self.app_id})"

    def __eq__(self, other_obj: object) -> bool:
        if not isinstance(other_obj, Farm):
            return False
        return self.app_id == other_obj.app_id

    @property
    def suggested_params(self) -> algosdk.transaction.SuggestedParams:
        assert self._suggested_params is not None
        return self._suggested_params

    def set_suggested_params(
        self, suggested_params: algosdk.transaction.SuggestedParams
    ):
        self._suggested_params = suggested_params

    def refresh_suggested_params(self):
        self.set_suggested_params(self.algod.suggested_params())

    def fetch_all_assets(self):
        self.state.staked_asset = fetch_asset_by_index(
            self.algod, self.state.staked_asset.index
        )
        self.state.reward_assets = [
            fetch_asset_by_index(self.algod, asset.index)
            for asset in self.state.reward_assets
        ]

    @property
    def staked_asset(self):
        return self.state.staked_asset

    def have_rewards(self, dt: Optional[datetime.datetime] = None) -> bool:
        state = self.state

        if state.duration == 0:
            # Finished distributing rewards or there never were any rewards.
            return False

        if state.total_staked == 0:
            # The farm is paused and still has rewards.
            return True

        if dt is None:
            dt = datetime.datetime.now()

        duration = datetime.timedelta(seconds=state.duration + state.next_duration)
        if dt < state.updated_at + duration:
            # The farm is going and still has rewards.
            return True

        # All the rewards will be distributed at the specified time.
        return False

    def fetch_escrow_by_id(self, app_id: int) -> Escrow:
        return fetch_escrow_by_id(self.algod, app_id, farm=self)

    def fetch_escrow_by_address(self, address: str) -> Optional[Escrow]:
        user_state = self.fetch_user_state(address)
        if user_state is None:
            return None
        return fetch_escrow_by_id(self.algod, user_state.escrow_id, farm=self)

    def fetch_escrow_from_account_info(self, account_info: dict) -> Optional[Escrow]:
        user_state = self.get_user_state_from_account_info(account_info)
        if user_state is None:
            return None
        return fetch_escrow_by_id(self.algod, user_state.escrow_id, farm=self)

    def update_state(self):
        app_info = self.algod.application_info(self.app_id)
        self.raw_state = parse_app_state(app_info["params"]["global-state"])
        self.internal_state = parse_internal_state(self.raw_state)
        self.state = internal_state_to_state(self.algod, self.internal_state)

    def fetch_user_state(self, address: str) -> Optional[FarmUserState]:
        account_info = self.algod.account_info(address)
        return self.get_user_state_from_account_info(account_info)

    def get_user_state_from_account_info(
        self, account_info: dict
    ) -> Optional[FarmUserState]:
        try:
            raw_state = parse_app_state(
                next(
                    state["key-value"]
                    for state in account_info["apps-local-state"]
                    if state["id"] == self.app_id
                )
            )
        except StopIteration:
            return None

        rpt = format_rpt(
            deserialize_uint64(raw_state["RPT"]),
            deserialize_uint64(raw_state["RPT_frac"]),
        )

        return FarmUserState(
            escrow_id=raw_state["EscrowID"],
            staked=raw_state["Staked"],
            accrued_rewards=format_rewards(
                self.state.reward_assets,
                deserialize_uint64(raw_state["AccruedRewards"]),
            ),
            claimed_rewards=format_rewards(
                self.state.reward_assets,
                deserialize_uint64(raw_state["ClaimedRewards"]),
            ),
            rpt=format_rewards(self.state.reward_assets, rpt),
        )

    def estimate_accrued_rewards(
        self,
        at_time: datetime.datetime,
        user_state: FarmUserState,
    ) -> FarmingRewards[int]:
        past_accrued_rewards = self._calculate_past_accrued_rewards(
            staked_amount=user_state.staked, user_rpt=user_state.rpt
        )

        estimated_rewards = self.simulate_accrued_rewards(
            at_time=at_time,
            staked_amount=user_state.staked,
            total_staked=self.state.total_staked,
        )

        rewards = self._sum_rewards(estimated_rewards, user_state.accrued_rewards)
        return self._sum_rewards(rewards, past_accrued_rewards)

    def simulate_new_staker(
        self, at_time: datetime.datetime, staked_amount: int
    ) -> FarmingRewards[int]:
        return self.simulate_accrued_rewards(
            at_time=at_time,
            staked_amount=staked_amount,
            total_staked=self.state.total_staked + staked_amount,
            extrapolate_future_rewards=True,
        )

    def simulate_accrued_rewards(
        self,
        at_time: datetime.datetime,
        staked_amount: int,
        total_staked: int,
        extrapolate_future_rewards=False,
    ) -> FarmingRewards[int]:
        duration = int((at_time - self.state.updated_at).total_seconds())

        if total_staked == 0:
            return {asset: 0 for asset in self.state.reward_assets}

        stake_ratio = staked_amount / total_staked

        # Simulate pending rewards.
        rewards = self._simulate_cycle_rewards(
            stake_ratio=stake_ratio,
            rewards=self.state.pending_rewards,
            stake_duration=duration,
            cycle_duration=self.state.duration,
        )

        duration -= self.state.duration
        if duration <= 0:
            return rewards

        if self.state.next_duration:
            # Simulate next rewards if needed.
            rewards_b = self._simulate_cycle_rewards(
                stake_ratio=stake_ratio,
                rewards=self.state.next_rewards,
                stake_duration=duration,
                cycle_duration=self.state.next_duration,
            )
            rewards = self._sum_rewards(rewards, rewards_b)

            duration -= self.state.next_duration
            if duration <= 0:
                return rewards

        if not extrapolate_future_rewards:
            return rewards

        next_duration = self.state.next_duration or self.state.duration
        if next_duration == 0:
            return rewards

        next_rewards = (
            self.state.next_rewards
            if self.state.next_duration
            else self.state.pending_rewards
        )
        next_next_rewards = {
            asset: int(amount * (duration / next_duration))
            for asset, amount in next_rewards.items()
        }

        # Extrapolate rewards for the future.
        rewards_c = self._simulate_cycle_rewards(
            stake_ratio=stake_ratio,
            rewards=next_next_rewards,
            stake_duration=duration,
            cycle_duration=duration,
        )

        return self._sum_rewards(rewards, rewards_c)

    def _simulate_cycle_rewards(
        self,
        stake_ratio: float,
        rewards: FarmingRewards[int],
        stake_duration: int,
        cycle_duration: int,
    ) -> FarmingRewards[int]:
        if cycle_duration == 0:
            return {asset: 0 for asset in self.state.reward_assets}

        stake_duration = min(stake_duration, cycle_duration)

        return {
            asset: int(
                stake_ratio * rewards.get(asset, 0) * (stake_duration / cycle_duration)
            )
            for asset in self.state.reward_assets
        }

    def _calculate_past_accrued_rewards(
        self,
        staked_amount: int,
        user_rpt: FarmingRewards[float],
    ) -> FarmingRewards[int]:
        return {
            asset: int(
                (
                    max(0, self.state.rpt.get(asset, 0) - user_rpt.get(asset, 0))
                    * staked_amount
                )
            )
            for asset in self.state.reward_assets
        }

    def _sum_rewards(
        self, rewards_a: FarmingRewards[int], rewards_b: FarmingRewards[int]
    ) -> FarmingRewards[int]:
        return {
            asset: amount + rewards_b.get(asset, 0)
            for asset, amount in rewards_a.items()
        }

    def prepare_deploy_escrow_txs(self, sender: str) -> list[transaction.Transaction]:
        return build_deploy_escrow_txs(
            sender=sender,
            farm_app_id=self.app_id,
            staked_asset_id=self.staked_asset.index,
            suggested_params=self.suggested_params,
        )

    def build_update_increase_opcode_quota_tx(
        self, sender: str
    ) -> Optional[transaction.Transaction]:
        seconds_passed = (datetime.datetime.now() - self.state.updated_at).seconds
        opcodes_cost = 671 if seconds_passed > self.state.duration > 0 else 513
        count = (opcodes_cost * len(self.state.reward_assets)) // 700
        if count == 0:
            return None

        return get_gas_station().build_increase_opcode_quota_tx(
            sender=sender,
            count=count,
            suggested_params=self.suggested_params,
        )

    def build_update_with_opcode_increase_txs(
        self, escrow: Escrow
    ) -> list[transaction.Transaction]:
        txs = [self.build_update_tx(escrow)]

        if increase_opcode_quota_tx := self.build_update_increase_opcode_quota_tx(
            escrow.user_address
        ):
            txs.insert(0, increase_opcode_quota_tx)

        return txs

    def build_update_tx(self, escrow: Escrow) -> transaction.Transaction:
        return transaction.ApplicationNoOpTxn(
            sender=escrow.user_address,
            index=self.app_id,
            foreign_assets=[self.staked_asset.index],
            foreign_apps=[escrow.app_id],
            accounts=[escrow.address],
            app_args=[
                UPDATE_STATE_SIG,
                abi.UintType(8).encode(1),
                abi.UintType(8).encode(1),
                abi.UintType(8).encode(0),
                abi.UintType(8).encode(0),
            ],
            sp=sp_fee(self.suggested_params, UPDATE_TX_FEE),
        )

    def build_claim_rewards_tx(
        self, escrow: Escrow, assets: Optional[list[Asset]] = None
    ) -> transaction.Transaction:
        if assets is None:
            assets = self.state.reward_assets

        number_of_assets = len(assets)
        return transaction.ApplicationNoOpTxn(
            sender=escrow.user_address,
            index=self.app_id,
            foreign_assets=[asset.index for asset in assets],
            foreign_apps=[escrow.app_id],
            accounts=[escrow.user_address],
            app_args=[
                CLAIM_REWARDS_SIG,
                abi.UintType(8).encode(0),
                abi.ArrayDynamicType(abi.UintType(64)).encode(
                    [self.state.reward_assets.index(asset) for asset in assets]
                ),
            ],
            sp=sp_fee(self.suggested_params, 1000 * (number_of_assets + 1)),
        )

    def build_update_global_state_tx(self, sender: str):
        return transaction.ApplicationNoOpTxn(
            sender=sender,
            index=self.app_id,
            app_args=[UPDATE_GLOBAL_STATE_SIG],
            sp=self.suggested_params,
        )

    def admin_build_add_reward_asset_tx(self, asset: Asset) -> transaction.Transaction:
        return transaction.ApplicationNoOpTxn(
            sender=self.state.admin,
            sp=sp_fee(self.suggested_params, 2000),
            index=self.app_id,
            foreign_assets=[asset.index],
            app_args=[ADD_REWARD_ASSET_SIG, 0],
        )

    def admin_build_deposit_rewards_txs(
        self, rewards: FarmingRewards[int], duration: int
    ) -> list[transaction.Transaction]:
        assets_to_opt_in = [
            asset for asset in rewards.keys() if asset not in self.state.reward_assets
        ]

        assert (
            self.state.next_duration == 0
        ), "Cannot deposit next rewards if farm already have next rewards"

        assert (
            len(self.state.reward_assets) + len(assets_to_opt_in) <= MAX_REWARD_ASSETS
        ), f"Maximum number of reward assets per farm is {MAX_REWARD_ASSETS}"

        increase_opcode_tx = None
        if self.state.total_staked:
            increase_opcode_tx = self.build_update_increase_opcode_quota_tx(
                self.state.admin
            )

        update_tx = self.build_update_global_state_tx(sender=self.state.admin)

        # Fund farm with minimal ALGO balance required for assets opt-ins.
        opt_in_txs: list[transaction.Transaction] = [
            transaction.PaymentTxn(
                sender=self.state.admin,
                receiver=self.app_address,
                amt=len(assets_to_opt_in) * 100_000,
                sp=self.suggested_params,
            )
        ]

        if assets_to_opt_in:
            opt_in_txs += [
                self.admin_build_add_reward_asset_tx(asset)
                for asset in assets_to_opt_in
            ]

        for asset in assets_to_opt_in:
            self.state.reward_assets.append(asset)

        transfer_txs = [
            asset.build_transfer_tx(
                sender=self.state.admin,
                receiver=self.app_address,
                amount=amount,
                suggested_params=self.suggested_params,
            )
            for asset, amount in rewards.items()
        ]

        foreign_assets = [asset for asset in rewards.keys()]

        app_args = [
            DEPOSIT_REWARDS_SIG,
            abi.ArrayDynamicType(abi.UintType(64)).encode(
                [self.state.reward_assets.index(asset) for asset in foreign_assets]
            ),
            abi.UintType(64).encode(duration),
        ]

        deposit_rewards_tx = transaction.ApplicationNoOpTxn(
            sender=self.state.admin,
            sp=self.suggested_params,
            index=self.app_id,
            app_args=app_args,
        )

        txs = [
            increase_opcode_tx,
            update_tx,
            *opt_in_txs,
            *transfer_txs,
            deposit_rewards_tx,
        ]

        return [tx for tx in txs if tx]



import dataclasses
from typing import Literal

MAINNET_API_URL = "https://api.pact.fi"
MAINNET_GAS_STATION_ID = 1027956681
MAINNET_FACTORY_CONSTANT_PRODUCT_ID = 1072843805
MAINNET_FACTORY_NFT_CONSTANT_PRODUCT_ID = 1076423760

TESTNET_API_URL = "https://api.testnet.pact.fi"
TESTNET_GAS_STATION_ID = 156575978
TESTNET_FACTORY_CONSTANT_PRODUCT_ID = 166540424
TESTNET_FACTORY_NFT_CONSTANT_PRODUCT_ID = 166540708

Network = Literal["mainnet", "testnet", "dev"]


@dataclasses.dataclass
class Config:
    api_url: str = ""
    gas_station_id: int = 0
    factory_constant_product_id: int = 0
    factory_nft_constant_product_id: int = 0


def get_config(network: Network, **kwargs) -> Config:
    if network == "mainnet":
        params: dict = {
            "api_url": MAINNET_API_URL,
            "gas_station_id": MAINNET_GAS_STATION_ID,
            "factory_constant_product_id": MAINNET_FACTORY_CONSTANT_PRODUCT_ID,
            "factory_nft_constant_product_id": MAINNET_FACTORY_NFT_CONSTANT_PRODUCT_ID,
            **kwargs,
        }
    elif network == "testnet":
        params = {
            "api_url": TESTNET_API_URL,
            "gas_station_id": TESTNET_GAS_STATION_ID,
            "factory_constant_product_id": TESTNET_FACTORY_CONSTANT_PRODUCT_ID,
            "factory_nft_constant_product_id": TESTNET_FACTORY_NFT_CONSTANT_PRODUCT_ID,
            **kwargs,
        }
    elif network == "dev":
        params = {
            "api_url": "",
            "gas_station_id": 0,
            "factory_constant_product_id": 0,
            "factory_nft_constant_product_id": 0,
            **kwargs,
        }
    else:
        raise ValueError(f"No predefined config for network {network}")

    return Config(**params)



import base64

from algosdk import transaction

from pactsdk.exceptions import PactSdkError


class TransactionGroup:
    """A convenience class to make managing Algorand transactions groups easier."""

    transactions: list[transaction.Transaction]
    """A list of transactions in a group."""

    def __init__(self, transactions: list[transaction.Transaction]):
        """Creates the TransactionGroup from an array of transactions by assigning a group id to each transaction.

        Args:
            transactions: A list of transactions to put in a group.

        Raises:
            PactSdkError: If the list is empty (length 0) or if the group id was not assigned to the transactions due to e.g. a failure in the Algorand SDK.
        """
        if len(transactions) == 0:
            raise PactSdkError(
                "Cannot create TransactionGroup: empty transactions list."
            )
        self.transactions = transaction.assign_group_id(transactions)

        first_tx = self.transactions[0]
        if not first_tx.group:
            raise PactSdkError("Cannot retrieve group id from transaction.")
        self.group_id_buffer = first_tx.group

    def sign(self, private_key: str) -> list[transaction.SignedTransaction]:
        """Signs all the transactions in the group with the private key.

        Using the private_key passed in assigns all the transactions stored internally for the transaction group.

        Args:
            private_key: Sign the transactions with this private key.

        Returns:
            A list of encoded signed transactions as per the Transaction.sign from the Algorand sdk.
        """
        return [tx.sign(private_key) for tx in self.transactions]

    @property
    def group_id(self):
        """
        Returns
            The group id as a base64 encoded string.
        """
        return base64.b64encode(self.group_id_buffer)



from .escrow import (  # noqa
    Escrow,
    EscrowInternalState,
    build_deploy_escrow_txs,
    fetch_escrow_approval_program,
    fetch_escrow_by_id,
    fetch_escrow_global_state,
    parse_global_escrow_state,
)
from .farm import (  # noqa
    Farm,
    fetch_farm_by_id,
    fetch_farm_raw_state_by_id,
    make_farm_from_raw_state,
)
from .farm_state import (  # noqa
    FarmingRewards,
    FarmInternalState,
    FarmState,
    FarmUserState,
    internal_state_to_state,
    parse_internal_state,
)
from .farming_client import PactFarmingClient  # noqa



"""Set of utility classes for managing and performing zaps.
"""
import copy
import math
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from .add_liquidity import LiquidityAddition
from .asset import Asset
from .swap import Swap
from .transaction_group import TransactionGroup

if TYPE_CHECKING:
    from .pool import Pool


FEE_PRECISION = 10**4


@dataclass
class ZapParams:
    """All amounts that should be used in swap and add liquidity transactions."""

    swap_deposited: int
    primary_add_liq: int
    secondary_add_liq: int


def get_constant_product_zap_params(
    liq_a: int,
    liq_b: int,
    zap_amount: int,
    fee_bps: int,
    pact_fee_bps: int,
) -> ZapParams:
    swap_deposited = get_swap_amount_deposited_from_zapping(
        zap_amount, liq_a, fee_bps, pact_fee_bps
    )
    primary_add_liq = zap_amount - swap_deposited
    secondary_add_liq = get_secondary_added_liquidity_from_zapping(
        swap_deposited, liq_a, liq_b, fee_bps
    )
    return ZapParams(swap_deposited, primary_add_liq, secondary_add_liq)


def get_swap_amount_deposited_from_zapping(
    zap_amount: int,
    total_amount: int,
    fee_bps: int,
    pact_fee_bps: int,
) -> int:
    pool_fee = fee_bps - pact_fee_bps
    a = (-FEE_PRECISION - pool_fee + fee_bps) // FEE_PRECISION
    b = (
        -2 * total_amount * FEE_PRECISION
        + zap_amount * pool_fee
        + total_amount * fee_bps
    ) // FEE_PRECISION
    c = total_amount * zap_amount

    delta = b * b - 4 * a * c
    if b < 0:
        result = (-b - math.isqrt(delta)) // (2 * a)
    else:
        result = (2 * c) // (-b + math.isqrt(delta))
    return result


def get_secondary_added_liquidity_from_zapping(
    swap_deposited: int,
    total_primary: int,
    total_secondary: int,
    fee_bps,
) -> int:
    return (swap_deposited * total_secondary * (FEE_PRECISION - fee_bps)) // (
        (total_primary + swap_deposited) * FEE_PRECISION
    )


@dataclass
class Zap:
    """Zap class represents a zap trade on a particular pool, which allows to exchange single asset for PLP token.

    Zap performs a swap to get second asset from the pool and then adds liquidity using both of those assets. Users may be left with some leftovers due to rounding and slippage settings.

    Zaps are meant only for Constant Product pools; For Stableswaps, adding only one asset works out of the box.

    Typically, users don't have to manually instantiate this class. Use :py:meth:`pactsdk.pool.Pool.prepare_zap` instead.
    """

    pool: "Pool"
    """The pool the zap is going to be performed in."""

    asset: Asset
    """The asset that will be used in zap."""

    amount: int
    """Amount to be used in zap."""

    slippage_pct: float
    """The maximum amount of slippage allowed in performing the swap."""

    swap: Swap = field(init=False)
    """The swap object that will be executed during the zap."""

    liquidity_addition: LiquidityAddition = field(init=False)
    """Liquidity Addition object that will be executed during the zap."""

    params: ZapParams = field(init=False)
    """All amounts used in swap and add liquidity transactions."""

    def __post_init__(self):
        if not self.pool.is_asset_in_the_pool(self.asset):
            raise AssertionError("Provided asset was not found in the pool.")

        if self.pool.pool_type == "STABLESWAP":
            raise AssertionError("Zap can only be made on constant product pools.")

        self.params = self.get_zap_params()
        self.swap = self.pool.prepare_swap(
            self.asset, self.params.swap_deposited, self.slippage_pct
        )
        self.liquidity_addition = self.prepare_add_liq()

    def prepare_tx_group(self, address: str) -> TransactionGroup:
        """Creates the transactions needed to perform zap and returns them as a transaction group ready to be signed and committed.

        Args:
            address: The account that will be performing the zap.

        Returns:
            A transaction group that when executed will perform the zap.
        """
        return self.pool.prepare_zap_tx_group(self, address)

    def _is_asset_primary(self):
        return self.asset.index == self.pool.primary_asset.index

    def get_zap_params(self) -> ZapParams:
        A, B = self.pool.calculator.get_liquidities(self.asset)
        if A == 0 or B == 0:
            raise ValueError("Cannot create a Zap on empty pool.")

        params = get_constant_product_zap_params(
            A, B, self.amount, self.pool.params.fee_bps, self.pool.params.pact_fee_bps
        )
        if not self._is_asset_primary():
            temp = params.primary_add_liq
            params.primary_add_liq = params.secondary_add_liq - 1
            params.secondary_add_liq = temp
        else:
            params.secondary_add_liq -= 1

        return params

    def prepare_add_liq(self):
        updated_state = self.pool.state
        if self._is_asset_primary():
            updated_state.total_primary += self.params.swap_deposited
            updated_state.total_secondary -= self.params.secondary_add_liq
        else:
            updated_state.total_primary -= self.params.primary_add_liq
            updated_state.total_secondary += self.params.swap_deposited

        pool = copy.copy(self.pool)
        pool.state = updated_state
        return LiquidityAddition(
            pool=pool,
            primary_asset_amount=self.params.primary_add_liq,
            secondary_asset_amount=self.params.secondary_add_liq,
        )



"""Set of utility classes for managing and performing swaps.
"""
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from pactsdk.stableswap_calculator import StableswapCalculator, get_tx_fee

from .asset import Asset
from .transaction_group import TransactionGroup

if TYPE_CHECKING:
    from .pool import Pool


@dataclass
class SwapEffect:
    """Swap Effect are the basic details of the effect on the pool of performing the swap."""

    amount_received: int
    amount_deposited: int
    minimum_amount_received: int
    fee: int
    primary_asset_price_after_swap: float
    secondary_asset_price_after_swap: float
    primary_asset_price_change_pct: float
    secondary_asset_price_change_pct: float
    price: float
    tx_fee: int

    amplifier: float
    """Stableswap only. Zero otherwise."""


@dataclass
class Swap:
    """Swap class represents a swap trade on a particular pool.

    Typically, users don't have to manually instantiate this class. Use :py:meth:`pactsdk.pool.Pool.prepare_swap` instead.
    """

    pool: "Pool"
    """The pool the swap is going to be performed in."""

    asset_deposited: Asset
    """The asset that will be swapped (deposited in the contract)."""

    asset_received: Asset = field(init=False)
    """The asset that will be received."""

    amount: int
    """Either the amount to swap (deposit) or the amount to receive depending on the `swap_for_exact` parameter."""

    slippage_pct: float
    """The maximum percentage of slippage allowed in performing the swap."""

    swap_for_exact: bool = False
    """If `true` then `amount` is what you want to receive from the swap. Otherwise, it's an amount that you want to swap (deposit). Note that the contracts do not support the "swap exact for" swap. It works by calculating the amount to deposit on the client side and doing a normal swap on the exchange."""

    effect: SwapEffect = field(init=False)
    """The effect of the swap computed at the time of construction."""

    def __post_init__(self):
        self.asset_received = self.pool.get_other_asset(self.asset_deposited)
        self._validate_swap()
        self.effect = self._build_effect()

    def prepare_tx_group(self, address: str) -> TransactionGroup:
        """Creates the transactions needed to perform the swap trade and returns them as a transaction group ready to be signed and committed.

        Args:
            address: The account that will be performing the swap.

        Returns:
            A transaction group that when executed will perform the swap.
        """
        return self.pool.prepare_swap_tx_group(self, address)

    def _validate_swap(self):
        if self.slippage_pct < 0 or self.slippage_pct > 100:
            raise ValueError("Splippage must be between 0 and 100")

        if self.pool.calculator.is_empty:
            raise ValueError("Pool is empty and swaps are impossible.")

    def _build_effect(self) -> SwapEffect:
        if self.swap_for_exact:
            amount_received = self.amount
            amount_deposited = (
                self.pool.calculator.net_amount_received_to_amount_deposited(
                    self.asset_deposited, self.amount
                )
            )
        else:
            amount_received = (
                self.pool.calculator.amount_deposited_to_net_amount_received(
                    self.asset_deposited, self.amount
                )
            )
            amount_deposited = self.amount

        if self.asset_deposited == self.pool.primary_asset:
            primary_liq_change = amount_deposited
            secondary_liq_change = -amount_received
        else:
            primary_liq_change = -amount_received
            secondary_liq_change = amount_deposited

        amplifier = 0
        tx_fee = 2000
        swap_calc = self.pool.calculator.swap_calculator

        if isinstance(swap_calc, StableswapCalculator):
            amplifier = swap_calc.get_amplifier() / (
                self.pool.internal_state.PRECISION or 1
            )
            tx_fee = get_tx_fee(swap_calc.swap_invariant_iterations, 1)

        return SwapEffect(
            amount_deposited=amount_deposited,
            amount_received=amount_received,
            minimum_amount_received=self.pool.calculator.get_minimum_amount_received(
                self.asset_deposited, amount_deposited, self.slippage_pct
            ),
            price=self.pool.calculator.get_swap_price(
                self.asset_deposited, amount_deposited
            ),
            primary_asset_price_after_swap=self.pool.calculator.get_asset_price_after_liq_change(
                self.pool.primary_asset,
                primary_liq_change,
                secondary_liq_change,
            ),
            secondary_asset_price_after_swap=self.pool.calculator.get_asset_price_after_liq_change(
                self.pool.secondary_asset,
                primary_liq_change,
                secondary_liq_change,
            ),
            primary_asset_price_change_pct=self.pool.calculator.get_price_impact_pct(
                self.pool.primary_asset,
                primary_liq_change,
                secondary_liq_change,
            ),
            secondary_asset_price_change_pct=self.pool.calculator.get_price_impact_pct(
                self.pool.secondary_asset,
                primary_liq_change,
                secondary_liq_change,
            ),
            fee=self.pool.calculator.get_fee(self.asset_deposited, amount_deposited),
            amplifier=amplifier,
            tx_fee=tx_fee,
        )



"""Module for interacting with Pact API."""

from typing import TypedDict
from urllib.parse import urlencode

import requests


class ListPoolsParams(TypedDict, total=False):
    """Parameters for calling the :py:func:`pactsdk.pool.list_pools` function.

    All keys are optional and can be omitted.
    """

    offset: int
    limit: int
    is_verified: str
    creator: str
    primary_asset__algoid: int
    secondary_asset__algoid: int
    primary_asset__unit_name: str
    secondary_asset__unit_name: str
    primary_asset__name: str
    secondary_asset__name: str


class ApiAsset(TypedDict):
    """Details about the liquidity pool assets returned from the asset pool."""

    algoid: str
    decimals: int
    id: int
    is_liquidity_token: bool
    is_verified: bool
    name: str
    total_amount: str
    tvl_usd: str
    unit_name: str
    volume_7d: str
    volume_24h: str


class ApiPool(TypedDict):
    """The individual pool information returned from :py:func:`pactsdk.pool.list_pools`, this contains the basic pool information."""

    address: str
    appid: str
    confirmed_round: int
    creator: str
    fee_amount_7d: str
    fee_amount_24h: str
    fee_usd_7d: str
    fee_usd_24h: str
    tvl_usd: str
    volume_7d: str
    volume_24h: str
    apr_7d: str
    id: int
    is_verified: bool
    pool_asset: ApiAsset
    primary_asset: ApiAsset
    secondary_asset: ApiAsset


class ApiListPoolsResponse(TypedDict):
    """Response from :py:func:`pactsdk.pool.list_pools` function containing pagination information and results."""

    count: int
    offset: int
    limit: int
    results: list[ApiPool]


def list_pools(pact_api_url: str, params: ListPoolsParams) -> ApiListPoolsResponse:
    """Finds all the pools that match the pool options passed in.

    Args:
        pact_api_url: The API URL to query the list of pools.
        params: Dict of params for querying the pools.

    Returns:
        Pool data for all pools in the Pact that meets the pool options.
    """
    assert pact_api_url
    encoded_params = urlencode(params)
    response = requests.get(f"{pact_api_url}/api/pools?{encoded_params}")
    return response.json()



from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal, Optional

from pactsdk.encoding import (
    decode_address_from_global_state,
    decode_string_from_global_state,
    deserialize_uint64,
)

from .utils import parse_app_state

if TYPE_CHECKING:
    from pactsdk.pool import PoolType


@dataclass
class AppInternalState:
    """The one to one representation of pool's global state."""

    # Properties shared for all contracts.
    A: int
    B: int
    ASSET_A: int
    ASSET_B: int
    LTID: int
    L: int
    FEE_BPS: int

    # Those may be missing in older contracts.
    CONTRACT_NAME: Optional[Literal["PACT AMM", "[SI] PACT AMM"]] = None
    VERSION: Optional[int] = None
    PACT_FEE_BPS: Optional[int] = None
    ADMIN: Optional[str] = None
    FUTURE_ADMIN: Optional[str] = None
    TREASURY: Optional[str] = None
    PRIMARY_FEES: Optional[int] = None
    SECONDARY_FEES: Optional[int] = None

    # Stableswaps only below.
    INITIAL_A: Optional[int] = None
    INITIAL_A_TIME: Optional[int] = None
    FUTURE_A: Optional[int] = None
    FUTURE_A_TIME: Optional[int] = None
    PRECISION: Optional[int] = None


@dataclass
class PoolState:
    """A user friendly representation of pool's global state."""

    total_liquidity: int
    total_primary: int
    total_secondary: int
    primary_asset_price: float
    secondary_asset_price: float


def parse_global_pool_state(raw_state: list) -> AppInternalState:
    """
    Args:
        raw_state: The contract's global state retrieved from algosdk.
    """
    state = parse_app_state(raw_state)

    if "CONTRACT_NAME" in state:
        state["CONTRACT_NAME"] = decode_string_from_global_state(state["CONTRACT_NAME"])

    if "ADMIN" in state:
        state["ADMIN"] = decode_address_from_global_state(state["ADMIN"])

    if "TREASURY" in state:
        state["TREASURY"] = decode_address_from_global_state(state["TREASURY"])

    if "INITIAL_A" in state:
        asset_a, asset_b, _, precision = deserialize_uint64(state.pop("CONFIG"))

        return AppInternalState(
            ASSET_A=asset_a,
            ASSET_B=asset_b,
            PRECISION=precision,
            **state,
        )

    if "FEE_BPS" in state:
        del state["FEE_BPS"]

    asset_a, asset_b, fee_bps = deserialize_uint64(state.pop("CONFIG"))
    return AppInternalState(ASSET_A=asset_a, ASSET_B=asset_b, FEE_BPS=fee_bps, **state)


def get_pool_type_from_internal_state(
    state: AppInternalState,
) -> "PoolType":
    if state.CONTRACT_NAME == "PACT AMM":
        return "CONSTANT_PRODUCT"

    if state.CONTRACT_NAME == "PACT AMM [NFT]":
        return "NFT_CONSTANT_PRODUCT"

    if state.CONTRACT_NAME == "[SI] PACT AMM":
        return "STABLESWAP"

    # Older contracts are missing CONTRACT_NAME. Let's assume it's our good old constant product.
    return "CONSTANT_PRODUCT"



import base64
from copy import copy
from typing import Any

import algosdk
from Cryptodome.Hash import SHA512


def sp_fee(
    sp: algosdk.transaction.SuggestedParams, fee: int
) -> algosdk.transaction.SuggestedParams:
    """
    Get a copy of suggested params but with a different fee.
    """
    sp = copy(sp)
    sp.flat_fee = True
    sp.fee = fee
    return sp


def parse_app_state(kv: list) -> dict[str, Any]:
    """Utility function for converting the Algorand key-value schema into a python dictionary.

    Algorand store keys in base64 encoding and store values as either bytes or unsigned integers depending on the type. This function decodes this information into a more human friendly structure.

    Args:
        kv: Algorand key-value data structure to parse.

    Returns:
        The parsed key value dictionary.
    """
    res = {}
    for elem in kv:
        key = str(base64.b64decode(elem["key"]), encoding="ascii")
        if elem["value"]["type"] == 1:
            val = elem["value"]["bytes"]
        else:
            val = elem["value"]["uint"]
        res[key] = val
    return res


def get_selector(method_signature: str) -> bytes:
    hash_ = SHA512.new(truncate="256")
    hash_.update(method_signature.encode("utf-8"))
    return hash_.digest()[:4]


def get_box_min_balance(len_box_name: int, box_size: int) -> int:
    # https://developer.algorand.org/articles/smart-contract-storage-boxes/
    assert len_box_name <= 64
    return 2500 + 400 * (len_box_name + box_size)


def get_last_round(algod: algosdk.v2client.algod.AlgodClient) -> int:
    return algod.status().get("last-round")


# Function from Algorand Inc.
def wait_for_confirmation(algod: algosdk.v2client.algod.AlgodClient, txid: str) -> dict:
    """
    Utility function to wait until the transaction is
    confirmed before proceeding.
    """
    last_round = get_last_round(algod)
    txinfo = algod.pending_transaction_info(txid)
    while not (txinfo.get("confirmed-round") and txinfo.get("confirmed-round") > 0):
        last_round += 1
        algod.status_after_block(last_round)
        txinfo = algod.pending_transaction_info(txid)
    return txinfo



__version__ = "0.6.1"

from .add_liquidity import LiquidityAddition  # noqa
from .asset import Asset, fetch_asset_by_index  # noqa
from .client import PactClient  # noqa
from .exceptions import PactSdkError  # noqa
from .factories import *  # noqa
from .farming import *  # noqa
from .gas_station import GasStation, get_gas_station, set_gas_station  # noqa
from .pool import Pool, PoolState  # noqa
from .swap import Swap, SwapEffect  # noqa
from .transaction_group import TransactionGroup  # noqa
from .zap import Zap, ZapParams  # noqa



import logging
import math
import time
from dataclasses import dataclass
from math import isqrt
from typing import TYPE_CHECKING, cast

from .constant_product_calculator import get_constant_product_minted_liquidity_tokens
from .exceptions import PactSdkError

if TYPE_CHECKING:
    from .pool import Pool

logger = logging.getLogger(__name__)

MAX_GET_PRICE_RETRIES = 5


class ConvergenceError(PactSdkError):
    pass


@dataclass
class StableswapParams:
    fee_bps: int
    pact_fee_bps: int
    initial_a: int
    initial_a_time: int
    future_a: int
    future_a_time: int
    precision: int


def get_tx_fee(invariant_iterations: int, extra_margin: int) -> int:
    """To calculate the pool invariant, a Newton-Raphson method is used in both - the SDK and the smart contract.

    Algorand has a limit of the number of operations available in a single app call. To increase the limit, an additional empty inner transaction have to be created. Each extra tx increases tx fee. This functions calculates the fee needed for a swap transaction.

    Args:
        invariant_iterations: Number of iterations of Newton-Raphson interpolation.
        extra_margin: Number of extra inner transactions needed in case of slippage.

    Returns:
        The required fee.

    """
    inner_tx_count = math.ceil((invariant_iterations * 369) / 700)
    # +1 - first obligatory inner tx
    # +1 - app call
    # +2 in total
    return (inner_tx_count + 2 + extra_margin) * 1000


def get_stableswap_minted_liquidity_tokens(
    added_primary: int,
    added_secondary: int,
    total_primary: int,
    total_secondary: int,
    total_liquidity: int,
    amplifier: int,
    precision: int,
    fee_bps: int,
) -> tuple[int, int]:
    """Returns a tuple of minted tokens and total Newton-Raphson iterations (needed for tx fee calculations)."""
    if total_primary + total_secondary == 0:
        minted_tokens = get_constant_product_minted_liquidity_tokens(
            added_primary,
            added_secondary,
            total_primary,
            total_secondary,
            total_liquidity,
        )
        return minted_tokens, 0

    initial_totals = (total_primary, total_secondary)
    updated_totals = (
        total_primary + added_primary,
        total_secondary + added_secondary,
    )

    fees, initial_d, swap_invariant_iterations = get_add_liquidity_fees(
        initial_totals,
        updated_totals,
        fee_bps,
        amplifier,
        precision,
    )
    next_d, next_iterations = get_invariant(
        updated_totals[0] - fees[0],
        updated_totals[1] - fees[1],
        amplifier,
        precision,
    )

    minted_tokens = (total_liquidity * (next_d - initial_d)) // initial_d
    return minted_tokens, swap_invariant_iterations + next_iterations


def get_add_liquidity_bonus_pct(
    added_primary: int,
    added_secondary: int,
    total_primary: int,
    total_secondary: int,
    fee_bps: int,
    amplifier: int,
    precision: int,
) -> float:
    if total_primary + total_secondary == 0:
        return 0

    initial_totals = (total_primary, total_secondary)
    updated_totals = (
        total_primary + added_primary,
        total_secondary + added_secondary,
    )

    fees, initial_d, _ = get_add_liquidity_fees(
        initial_totals,
        updated_totals,
        fee_bps,
        amplifier,
        precision,
    )

    final_balances = (
        updated_totals[0] - fees[0],
        updated_totals[1] - fees[1],
    )

    final_d, _ = get_invariant(
        final_balances[0],
        final_balances[1],
        amplifier,
        precision,
    )

    # Calculate the gain in absolute terms, considering that each token is worth 1.
    total_added = added_primary + added_secondary
    return ((final_d - initial_d) / total_added - 1) * 100


def get_add_liquidity_fees(
    initial_totals: tuple[int, int],
    updated_totals: tuple[int, int],
    fee_bps: int,
    amplifier: int,
    precision: int,
) -> tuple[tuple[int, int], int, int]:
    n = 2

    initial_d, initial_iterations = get_invariant(
        initial_totals[0],
        initial_totals[1],
        amplifier,
        precision,
    )

    # Calculate the invariant as if all tokens were added to the pool
    next_d, next_iterations = get_invariant(
        updated_totals[0],
        updated_totals[1],
        amplifier,
        precision,
    )

    perfect_balances = [(next_d * total) // initial_d for total in initial_totals]
    deltas = (
        abs(updated_totals[0] - perfect_balances[0]),
        abs(updated_totals[1] - perfect_balances[1]),
    )

    fees = tuple((delta * fee_bps * n) // (10_000 * (4 * (n - 1))) for delta in deltas)

    return (fees[0], fees[1]), initial_d, initial_iterations + next_iterations


def get_swap_gross_amount_received(
    liq_a: int, liq_b: int, amount_deposited: int, amplifier: int, precision: int
) -> tuple[int, int]:
    invariant, iterations = get_invariant(liq_a, liq_b, amplifier, precision)
    new_liq_b = get_new_liq(liq_a + amount_deposited, amplifier, invariant, precision)
    return liq_b - new_liq_b, iterations


def get_swap_amount_deposited(
    liq_a: int,
    liq_b: int,
    gross_amount_received: int,
    amplifier: int,
    precision: int,
) -> tuple[int, int]:
    invariant, iterations = get_invariant(liq_a, liq_b, amplifier, precision)
    new_liq_a = get_new_liq(
        liq_b - gross_amount_received, amplifier, invariant, precision
    )
    return new_liq_a - liq_a, iterations


def get_invariant(liq_a: int, liq_b: int, amp: int, precision: int) -> tuple[int, int]:
    """Uses a Newton-Raphson method to calculate the pool invariant.

    Returns:
        A tuple of invariant and number of iterations required to calculate the invariant.
    """
    tokens_total = liq_a + liq_b
    S = tokens_total
    if S == 0:
        return S, 0

    D = S
    Ann = amp * 4
    i = 0
    while i < 64:
        i += 1
        D_P = D * D * D
        D_P //= liq_a * liq_b * 4

        Dprev = D
        numerator = D * ((Ann * S) // precision + D_P * 2)
        divisor = ((Ann - precision) * D) // precision + (2 + 1) * D_P
        D = numerator // divisor
        if D > Dprev:
            if D - Dprev <= 1:
                break
        elif Dprev - D <= 1:
            break
    if i == 64:
        raise ConvergenceError(f"Didn't converge {Dprev=}, {D=}")
    return D, i


def get_new_liq(liq_other: int, amplifier: int, inv: int, precision: int) -> int:
    S = liq_other
    D = inv
    A = amplifier
    P = liq_other
    Ann = A * 4

    b = S + (D * precision) // Ann
    c = (precision * (D**3)) // (4 * P * Ann)

    a_q = 1
    b_q = b - D
    c_q = -c

    delta = b_q * b_q - 4 * a_q * c_q
    return (-b_q + isqrt(delta)) // (2 * a_q)


def get_amplifier(
    timestamp: int,
    initial_a: int,
    initial_a_time: int,
    future_a: int,
    future_a_time: int,
) -> int:
    """Calculates the amplifier at specified timestamp using linear interpolation."""
    dt = future_a_time - initial_a_time
    dv = future_a - initial_a
    if not dt or not dv:
        return future_a

    dv_per_s = dv / dt

    min_a, max_a = (
        (initial_a, future_a) if future_a > initial_a else (future_a, initial_a)
    )

    current_a = initial_a + (timestamp - initial_a_time) * dv_per_s
    return max(min_a, min(max_a, math.ceil(current_a)))


class StableswapCalculator:
    """An implementation of a math behind stableswap pools."""

    swap_invariant_iterations = 0
    """Keeps the amount of iteration used to calculate invariant in the last call to getSwapGrossAmountReceived or getSwapAmountDeposited. Needed to calculate transaction fee."""

    mint_tokens_invariant_iterations = 0
    """The same as swap_invariant_iterations but for adding liquidity."""

    def __init__(self, pool: "Pool"):
        self.pool = pool

    @property
    def stableswap_params(self) -> StableswapParams:
        return cast(StableswapParams, self.pool.params)

    def get_amplifier(self):
        params = self.stableswap_params
        return get_amplifier(
            int(time.time()),
            params.initial_a,
            params.initial_a_time,
            params.future_a,
            params.future_a_time,
        )

    def get_price(self, liq_a: float, liq_b: float) -> float:
        """May return zero for highly unbalanced pools."""
        if not liq_a or not liq_b:
            return 0

        ratio = self.pool.primary_asset.ratio
        if ratio != self.pool.secondary_asset.ratio:
            logger.warning(
                "Number of decimals differs between primary and secondary asset. Stableswap does not support this scenario correctly.",
            )

        return self._get_price(liq_a, liq_b, MAX_GET_PRICE_RETRIES)

    def _get_price(self, liq_a: float, liq_b: float, retries: int) -> float:
        """
        Price is calculated by simulating a swap for 10**6 of micro values.
        This price is highly inaccurate for low liquidity pools.
        In case of ConvergenceError we try to simulate a swap using a different swap amount.
        Returns zero if all retries will fail.
        """
        if retries <= 0:
            return 0.0

        ratio = self.pool.primary_asset.ratio

        liq_a *= ratio
        liq_b *= ratio

        # The division helps minimize price impact of simulated swap.
        amount_deposited = 10 ** (6 + MAX_GET_PRICE_RETRIES - retries)
        amount_deposited = min(amount_deposited, int(liq_a // 100), int(liq_b // 100))

        try:
            amount_received = self.get_swap_gross_amount_received(
                int(liq_b), int(liq_a), int(amount_deposited), save_iterations=False
            )
            if amount_received == 0:
                return self._get_price(liq_a, liq_b, retries - 1)

            return amount_deposited / amount_received
        except ConvergenceError:
            return self._get_price(liq_a, liq_b, retries - 1)

    def get_swap_gross_amount_received(
        self,
        liq_a: int,
        liq_b: int,
        amount_deposited: int,
        save_iterations=True,
    ) -> int:
        amplifier = self.get_amplifier()
        precision = self.stableswap_params.precision
        amount_received, iterations = get_swap_gross_amount_received(
            liq_a, liq_b, amount_deposited, amplifier, precision
        )
        if save_iterations:
            self.swap_invariant_iterations = iterations
        return amount_received

    def get_swap_amount_deposited(
        self,
        liq_a: int,
        liq_b: int,
        gross_amount_received: int,
        save_iterations=True,
    ) -> int:
        amplifier = self.get_amplifier()
        precision = self.stableswap_params.precision
        amount_deposited, iterations = get_swap_amount_deposited(
            liq_a, liq_b, gross_amount_received, amplifier, precision
        )
        if save_iterations:
            self.swap_invariant_iterations = iterations
        return amount_deposited

    def get_minted_liquidity_tokens(self, added_liq_a: int, added_liq_b: int) -> int:
        precision = self.stableswap_params.precision
        amplifier = self.get_amplifier()

        minted_tokens, iterations = get_stableswap_minted_liquidity_tokens(
            added_liq_a,
            added_liq_b,
            self.pool.state.total_primary,
            self.pool.state.total_secondary,
            self.pool.state.total_liquidity,
            amplifier,
            precision,
            self.pool.fee_bps,
        )

        self.mint_tokens_invariant_iterations = iterations

        if minted_tokens > 0:
            return minted_tokens

        if minted_tokens == 0:
            raise PactSdkError(
                "Amount of minted liquidity tokens must be greater then 0.",
            )

        """
        Add liquidity fee is always taken from both assets, even if the user provided only one asset as the liquidity. In self case, the fee is taken from current pool's liquidity.
        If the current liquidity is not high enough to cover the fee, the contract will fail.
        In the SDK calculations self results in minted_tokens < 0.
        """
        raise PactSdkError(
            "Pool liquidity too low to cover add liquidity fee.",
        )



"""Utility functions and class for dealing with Algorand Standard Assets."""

import copy
from dataclasses import dataclass
from typing import Optional

from algosdk import transaction
from algosdk.v2client.algod import AlgodClient

ASSETS_CACHE: dict[tuple[AlgodClient, int], "Asset"] = {}
"""Dictionary mapping the asset index number to the :py:class:`pactsdk.asset.Asset` class to speed up look up of the asset information."""


def get_cached_asset(algod: AlgodClient, index: int, decimals: int) -> "Asset":
    cache_key = (algod, index)
    if cache_key in ASSETS_CACHE:
        return copy.copy(ASSETS_CACHE[cache_key])

    return Asset(algod=algod, index=index, decimals=decimals)


def fetch_asset_by_index(
    algod: AlgodClient,
    index: int,
) -> "Asset":
    """Fetches an :py:class:`pactsdk.asset.Asset` class with the details about the asset for a given id number.

    The function uses an internal cache so as to minimize the number of times the actual Algorand client is used to look up the asset. This function is used through out the pact sdk to query asset information.

    Args:
        algod: An Algorand client to query about the asset.
        index: An Algorand Asset number to look up.
    Returns:
        An asset instance for the index passed in.
    """
    cache_key = (algod, index)
    if cache_key in ASSETS_CACHE:
        return copy.copy(ASSETS_CACHE[cache_key])

    if index > 0:
        asset_info = algod.asset_info(index)
        params = asset_info["params"]
    else:
        params = {
            "name": "Algo",
            "unit-name": "ALGO",
            "decimals": 6,
        }

    asset = Asset(
        algod=algod,
        index=index,
        decimals=params["decimals"],
        name=params.get("name"),
        unit_name=params.get("unit-name"),
    )

    ASSETS_CACHE[cache_key] = asset

    return asset


@dataclass
class Asset:
    """Describes the basic data and the utility functions for an Algorand Standard Asset.

    Typically you don't create instances of this class manually. Use :py:meth:`pactsdk.client.PactClient.fetch_asset` instead. Also, when instantiating the pool e.g. by using :py:meth:`pactsdk.client.PactClient.fetch_pool_by_id` the missing pool assets are fetched automatically.
    """

    algod: AlgodClient
    """The Algorand sdk client to use for extracting asset details."""

    index: int
    """The ID of the asset."""

    decimals: int
    """The number of decimal places that the Asset supports."""

    name: Optional[str] = None
    """The name of the Asset if there is one. This may be None."""

    unit_name: Optional[str] = None
    """The name of a unit of the asset if there is one. This may be None."""

    @property
    def ratio(self) -> int:
        """The ratio between a base unit and the unit of the asset.

        This is used to convert between an integer and floating point representation of the asset without loss of precision.
        """
        return 10**self.decimals

    def prepare_opt_in_tx(self, address: str) -> transaction.AssetTransferTxn:
        """This creates a transaction that will allow the account to "opt in" to the asset.

        In Algorand, every account has to explicitly opt-in for an asset before receiving it.
        Needed if you want to receive an asset from a swap or to manage liquidity tokens.

        Args:
            address: Account to opt in to this asset.

        Returns:
            A ready to send transaction to opt-in into the ASA.
        """
        suggested_params = self.algod.suggested_params()
        return self.build_opt_in_tx(address, suggested_params)

    def build_opt_in_tx(
        self, address: str, suggested_params: transaction.SuggestedParams
    ) -> transaction.AssetTransferTxn:
        """Creates the actual transaction for the account to opt-in to holding the asset.

        Args:
            address: Address of the account to opt in to the asset.
            suggested_params: Algorand suggested parameters for transactions.

        Returns:
            A transaction to opt-in into asset.
        """
        return transaction.AssetTransferTxn(
            sender=address,
            receiver=address,
            amt=0,
            index=self.index,
            sp=suggested_params,
        )

    def prepare_opt_out_tx(
        self, address: str, close_to: str
    ) -> transaction.AssetTransferTxn:
        """This creates a transaction that will allow the account to "opt out" of the asset.

        Args:
            address: Account to opt out of this asset.

        Returns:
            A ready to send transaction to opt-out of the ASA.
        """
        suggested_params = self.algod.suggested_params()
        return self.build_opt_out_tx(address, close_to, suggested_params)

    def build_opt_out_tx(
        self, address: str, close_to: str, suggested_params: transaction.SuggestedParams
    ) -> transaction.AssetTransferTxn:
        """Creates the actual transaction for the account to opt-out from asset.

        Args:
            address: Address of the account to opt out of the asset.
            suggested_params: Algorand suggested parameters for transactions.

        Returns:
            A transaction to opt-out of the asset.
        """
        return transaction.AssetTransferTxn(
            sender=address,
            receiver=address,
            close_assets_to=close_to,
            amt=0,
            index=self.index,
            sp=suggested_params,
        )

    def is_opted_in(self, address: str) -> bool:
        """Checks if the account is already able to hold this asset, that is it has already opted in.

        This functions should be called to check if the opt-in transaction needs to be created. See :py:meth:`pactsdk.asset.Asset.prepare_opt_in_tx`.

        Args:
            address: The account to check if the asset is opted in on.

        Returns:
            True if the account is already opted in, false otherwise.
        """
        holding = self.get_holding(address)
        return holding is not None

    def get_holding(self, address: str) -> Optional[int]:
        """Returns the amount of holding of this asset the account has.

        Note that this function may return None if the account has not opted in for this asset.

        Args:
            address: The account to check the current holding.

        Returns:
            The amount of this asset the account is holding, or None if the account is not opted into the asset.
        """
        account_info = self.algod.account_info(address)
        return self.get_holding_from_account_info(account_info)

    def get_holding_from_account_info(self, account_info: dict) -> Optional[int]:
        if self.index == 0:
            return account_info["amount"]

        for asset in account_info["assets"]:
            if asset["asset-id"] == self.index:
                return asset["amount"]

        return None

    def build_transfer_tx(
        self,
        sender: str,
        receiver: str,
        amount: int,
        suggested_params: transaction.SuggestedParams,
        note: Optional[bytes] = None,
    ) -> transaction.Transaction:
        if self.index == 0:
            # ALGO
            return transaction.PaymentTxn(
                sender=sender,
                receiver=receiver,
                amt=amount,
                note=note,
                sp=suggested_params,
            )

        return transaction.AssetTransferTxn(
            sender=sender,
            receiver=receiver,
            amt=amount,
            note=note,
            sp=suggested_params,
            index=self.index,
        )

    def __eq__(self, other_asset: object) -> bool:
        """Return equal by comparing the assets index value."""
        if not isinstance(other_asset, Asset):
            return False
        return self.index == other_asset.index

    def __repr__(self):
        if self.unit_name:
            return f"<{self.unit_name}>"
        return f"<Asset {self.index}>"

    def __hash__(self) -> int:
        return self.index



import math
from dataclasses import dataclass
from typing import TYPE_CHECKING

from pactsdk.exceptions import PactSdkError

if TYPE_CHECKING:
    from .pool import Pool


@dataclass
class ConstantProductParams:
    fee_bps: int
    pact_fee_bps: int


def get_swap_gross_amount_received(
    liq_a: int, liq_b: int, amount_deposited: int
) -> int:
    return (liq_b * amount_deposited) // (liq_a + amount_deposited)


def get_swap_amount_deposited(
    liq_a: int,
    liq_b: int,
    gross_amount_received: int,
) -> int:
    deposited = liq_a * gross_amount_received / (liq_b - (gross_amount_received))
    return math.ceil(deposited)


def get_constant_product_minted_liquidity_tokens(
    added_primary: int,
    added_secondary: int,
    total_primary: int,
    total_secondary: int,
    total_liquidity: int,
) -> int:
    if total_primary + total_secondary == 0:
        return math.isqrt(added_primary * added_secondary)

    lt_a = (added_primary * total_liquidity) // total_primary
    lt_b = (added_secondary * total_liquidity) // total_secondary
    return lt_b if lt_a > lt_b else lt_a


class ConstantProductCalculator:
    """An implementation of a math behind constant product pools."""

    def __init__(self, pool: "Pool"):
        self.pool = pool

    def get_price(self, liq_a: float, liq_b: float) -> float:
        if not liq_a or not liq_b:
            return 0
        return liq_b / liq_a

    def get_swap_gross_amount_received(
        self, liq_a: int, liq_b: int, amount_deposited: int
    ) -> int:
        return get_swap_gross_amount_received(liq_a, liq_b, amount_deposited)

    def get_swap_amount_deposited(
        self,
        liq_a: int,
        liq_b: int,
        gross_amount_received: int,
    ) -> int:
        return get_swap_amount_deposited(liq_a, liq_b, gross_amount_received)

    def get_minted_liquidity_tokens(self, added_liq_a: int, added_liq_b: int) -> int:
        minted_tokens = get_constant_product_minted_liquidity_tokens(
            added_liq_a,
            added_liq_b,
            self.pool.state.total_primary,
            self.pool.state.total_secondary,
            self.pool.state.total_liquidity,
        )

        if minted_tokens > 0:
            return minted_tokens

        raise PactSdkError(
            "Amount of minted liquidity tokens must be greater then 0.",
        )



import datetime
from dataclasses import dataclass
from typing import TypeVar, Union

from algosdk.v2client.algod import AlgodClient

from pactsdk.asset import Asset

from ..asset import get_cached_asset
from ..encoding import decode_address_from_global_state, deserialize_uint64

T = TypeVar("T", bound=Union[int, float])

FarmingRewards = dict[Asset, T]


@dataclass()
class FarmInternalState:
    staked_asset_id: int
    reward_asset_ids: list[int]
    distributed_rewards: list[int]
    claimed_rewards: list[int]
    pending_rewards: list[int]
    next_rewards: list[int]
    rpt_frac: list[int]
    rpt: list[int]
    duration: int
    next_duration: int
    num_stakers: int
    total_staked: int
    updated_at: int
    admin: str
    updater: str
    version: int


@dataclass()
class FarmState:
    staked_asset: Asset
    """The asset the users are going to stake in the farm."""

    reward_assets: list[Asset]
    """Assets that are distributed as rewards in the farm."""

    distributed_rewards: FarmingRewards[int]
    """Amounts of assets distributed so far. This includes tokens that are already claimed and tokens which are accrued and are awaiting claim."""

    claimed_rewards: FarmingRewards[int]
    """Amounts of assets claimed by users so far."""

    pending_rewards: FarmingRewards[int]
    """Amounts of not yet distributed rewards in the farm."""

    rpt: FarmingRewards[float]
    """Current rate per token for each asset."""

    duration: int
    """Time in seconds until current cycle ends. This is the time at which the rewards are depleted. Next cycle is automatically picked up if next_rewards are deposited."""

    next_duration: int
    """The duration of the next cycle."""

    next_rewards: FarmingRewards[int]
    """Amounts of rewards deposited for the next cycle."""

    num_stakers: int
    """The number of active stakers. Active staker stakes at least 1 token."""

    total_staked: int
    """The sum of all stakers deposits."""

    updated_at: datetime.datetime
    """The time the farm was last updated."""

    admin: str
    """The address of the farm's admin account. The admin can deposit new rewards and destroy the farm after it is deprecated."""

    updater: str
    """The address of farm's updater. The updater can update the farm's contract to a new version."""

    version: int
    """Contract version."""

    @property
    def rewards_per_second(self) -> FarmingRewards[float]:
        return {
            asset: amount / self.duration
            for asset, amount in self.pending_rewards.items()
        }


@dataclass
class FarmUserState:
    escrow_id: int
    """The app id of the user's escrow contract."""

    staked: int
    """The amount of staked asset the user has deposited in the escrow."""

    accrued_rewards: FarmingRewards[int]
    """Amounts of rewards the user has accrued and can claim."""

    claimed_rewards: FarmingRewards[int]
    """Amounts of rewards the user has already claimed."""

    rpt: FarmingRewards[float]
    """Current rate per token for each asset."""


def parse_internal_state(raw_state: dict) -> FarmInternalState:
    return FarmInternalState(
        claimed_rewards=deserialize_uint64(raw_state["ClaimedRewards"]),
        duration=raw_state["Duration"],
        next_duration=raw_state["NextDuration"],
        next_rewards=deserialize_uint64(raw_state["NextRewards"]),
        num_stakers=raw_state["NumStakers"],
        pending_rewards=deserialize_uint64(raw_state["PendingRewards"]),
        rpt=deserialize_uint64(raw_state["RPT"]),
        rpt_frac=deserialize_uint64(raw_state["RPT_frac"]),
        reward_asset_ids=deserialize_uint64(raw_state["RewardAssetIDs"]),
        staked_asset_id=raw_state["StakedAssetID"],
        distributed_rewards=deserialize_uint64(raw_state["TotalRewards"]),
        total_staked=raw_state["TotalStaked"],
        updated_at=raw_state["UpdatedAt"],
        admin=decode_address_from_global_state(raw_state["Admin"]),
        updater=decode_address_from_global_state(raw_state["Updater"]),
        version=raw_state["VERSION"],
    )


def internal_state_to_state(
    algod: AlgodClient, internal_state: FarmInternalState
) -> FarmState:
    staked_asset = get_cached_asset(
        algod=algod, index=internal_state.staked_asset_id, decimals=0
    )

    reward_assets = [
        get_cached_asset(algod=algod, index=asset_id, decimals=0)
        for asset_id in internal_state.reward_asset_ids
    ]

    rpt = format_rpt(internal_state.rpt, internal_state.rpt_frac)

    return FarmState(
        staked_asset=staked_asset,
        reward_assets=reward_assets,
        distributed_rewards=format_rewards(
            reward_assets, internal_state.distributed_rewards
        ),
        claimed_rewards=format_rewards(reward_assets, internal_state.claimed_rewards),
        pending_rewards=format_rewards(reward_assets, internal_state.pending_rewards),
        next_rewards=format_rewards(reward_assets, internal_state.next_rewards),
        rpt=format_rewards(reward_assets, rpt),
        duration=internal_state.duration,
        next_duration=internal_state.next_duration,
        num_stakers=internal_state.num_stakers,
        total_staked=internal_state.total_staked,
        updated_at=datetime.datetime.fromtimestamp(internal_state.updated_at),
        admin=internal_state.admin,
        updater=internal_state.updater,
        version=internal_state.version,
    )


def format_rewards(assets: list[Asset], amounts: list[T]) -> FarmingRewards[T]:
    return {asset: amount for asset, amount in zip(assets, amounts)}


def format_rpt(rpt_whole: list[int], rpt_frac: list[int]) -> list[float]:
    assert len(rpt_whole) == len(rpt_frac)
    return [whole + frac / 2**64 for whole, frac in zip(rpt_whole, rpt_frac)]



"""Set of utility classes for adding liquidity to the pool.
"""


from dataclasses import dataclass, field
from typing import TYPE_CHECKING, cast

from pactsdk.exceptions import PactSdkError
from pactsdk.stableswap_calculator import (
    StableswapCalculator,
    StableswapParams,
    get_add_liquidity_bonus_pct,
    get_tx_fee,
)

from .constant_product_calculator import get_constant_product_minted_liquidity_tokens
from .transaction_group import TransactionGroup

if TYPE_CHECKING:
    from .pool import Pool


@dataclass
class AddLiquidityEffect:
    """The effect of adding liquidity to the pool."""

    minted_liquidity_tokens: int
    """Amount of new liquidity tokens minted when adding the liquidity. All the minted tokens will be received by the liquidity provider."""

    amplifier: float
    """Current stableswap amplifier. Zero for constant product pools."""

    bonus_pct: float
    """
    Only for stableswaps. Zero for constant product pools.

    A positive bonus means that after removing all the liquidity the user will end up with more tokens than he provided. This can happen when providing liquidity in a way that improves the pool balance.

    If adding the liquidity increases the pool imbalance, the bonus will be negative (a penalty).

    Also, a fee is subtracted from each liquidity addition. This negatively impacts the bonus.
    """

    tx_fee: int
    """App call transaction fee."""


@dataclass
class LiquidityAddition:
    """A representation of an action of adding liquidity to the pool.

    Typically, users don't have to manually instantiate this class. Use [[Pool.prepareAddLiquidity]] instead.
    """

    pool: "Pool"
    """The pool to provide liquidity for."""

    primary_asset_amount: int
    """Amount of primary asset the will be added to the pool."""

    secondary_asset_amount: int
    """Amount of secondary asset the will be added to the pool."""

    effect: AddLiquidityEffect = field(init=False)
    """The effect of adding the liquidity computed at the time of construction."""

    def __post_init__(self):
        self.effect = self.build_effect()

    def prepare_tx_group(self, address: str) -> TransactionGroup:
        """Creates the transactions needed to perform adding liquidity and returns them as a transaction group ready to be signed and committed.

        Args:
            address: The account that will be performing adding liquidity.

        Returns:
            A transaction group that when executed will add the liquidity to the pool.
        """
        return self.pool.prepare_add_liquidity_tx_group(
            address=address,
            liquidity_addition=self,
        )

    def build_effect(self) -> AddLiquidityEffect:
        amplifier = 0.0
        bonus_pct = 0.0
        tx_fee = 3000

        swap_calc = self.pool.calculator.swap_calculator
        state = self.pool.state

        if isinstance(swap_calc, StableswapCalculator):
            i_amplifier = swap_calc.get_amplifier()
            params = cast(StableswapParams, self.pool.params)
            bonus_pct = get_add_liquidity_bonus_pct(
                self.primary_asset_amount,
                self.secondary_asset_amount,
                state.total_primary,
                state.total_secondary,
                self.pool.fee_bps,
                i_amplifier,
                params.precision,
            )
            minted_liquidity_tokens = swap_calc.get_minted_liquidity_tokens(
                self.primary_asset_amount,
                self.secondary_asset_amount,
            )
            amplifier = i_amplifier / (self.pool.internal_state.PRECISION or 1)

            # 1 for each invariant calculation (3) and 1 for sending liquidity tokens.
            tx_fee = get_tx_fee(swap_calc.mint_tokens_invariant_iterations, 4)
        else:
            minted_liquidity_tokens = get_constant_product_minted_liquidity_tokens(
                self.primary_asset_amount,
                self.secondary_asset_amount,
                state.total_primary,
                state.total_secondary,
                state.total_liquidity,
            )
            if minted_liquidity_tokens <= 0:
                raise PactSdkError(
                    "Amount of minted liquidity tokens must be greater then 0.",
                )

        return AddLiquidityEffect(
            minted_liquidity_tokens=minted_liquidity_tokens,
            amplifier=amplifier,
            bonus_pct=bonus_pct,
            tx_fee=tx_fee,
        )



import base64
from typing import Sequence

import algosdk


# Not used but may be handy for e.g. users of the lib when writing unit tests.
def serialize_uint64(values: Sequence[int]) -> str:
    _bytes = bytes(
        x
        for i in values
        for x in int.to_bytes(i, length=8, byteorder="big", signed=False)
    )
    return base64.b64encode(_bytes).decode("ascii")


def deserialize_uint64(data: str) -> list[int]:
    decoded = base64.b64decode(data)
    return [
        int.from_bytes(decoded[offset : offset + 8], byteorder="big", signed=False)
        for offset in range(0, len(decoded), 8)
    ]


def decode_string_from_global_state(encoded: str) -> str:
    return base64.b64decode(encoded).decode()


def decode_address_from_global_state(encoded: str) -> str:
    return algosdk.encoding.encode_address(base64.b64decode(encoded))



"""
This module contains utilities for interacting with the escrow contract.
Each user deploys his own application which holds staked assets for only this user and for only one farm.
The contract is very minimal and has strong guarantees for user funds safety.
"""

import base64
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional

import algosdk
from algosdk import abi, transaction
from algosdk.v2client.algod import AlgodClient

from ..gas_station import get_gas_station
from ..utils import get_selector, parse_app_state, sp_fee

if TYPE_CHECKING:
    from .farm import Farm

COMPILED_HUSK_APPROVAL = "CCACAAExGEAASYALTWFzdGVyQXBwSUQ2GgEXwDJnsYEGshA2GgIXwDKyGIAEtzVf0bIaIrIBs7GBBLIQMgqyFCKyEjYaAxfAMLIRIrIBsyNCAAEjQw=="
COMPILED_CLEAR_PROGRAM_B64 = "CIEBQw=="

REKEY_TO_USER_FEE = 2000
REKEY_TO_CONTRACT_FEE = 1000

CREATE_SIG = get_selector("create(application,application,asset)void")
UNSTAKE_SIG = get_selector("unstake(asset,uint64,application)void")
SEND_MESSAGE_SIG = get_selector("send_message(account,string)void")
WITHDRAW_ALGOS_SIG = get_selector("withdraw_algos()void")


@dataclass
class EscrowInternalState:
    master_app: int


def fetch_escrow_approval_program(algod: AlgodClient, farm_app_id: int) -> bytes:
    box = algod.application_box_by_name(farm_app_id, b"Escrow")
    return base64.b64decode(box["value"])


def build_deploy_escrow_txs(
    sender: str,
    farm_app_id: int,
    staked_asset_id: int,
    suggested_params: transaction.SuggestedParams,
) -> list[transaction.Transaction]:

    approval_program = base64.b64decode(COMPILED_HUSK_APPROVAL)
    clear_program = base64.b64decode(COMPILED_CLEAR_PROGRAM_B64)

    gas_station = get_gas_station()

    fund_tx = gas_station.build_fund_tx(
        sender=sender,
        amount=200_000,
        suggested_params=suggested_params,
    )

    create_app_tx = transaction.ApplicationCreateTxn(
        sender=sender,
        approval_program=approval_program,
        clear_program=clear_program,
        on_complete=transaction.OnComplete.NoOpOC,
        global_schema=transaction.StateSchema(1, 0),
        local_schema=transaction.StateSchema(0, 0),
        sp=sp_fee(suggested_params, 5000),
        foreign_apps=[farm_app_id, gas_station.app_id],
        foreign_assets=[staked_asset_id],
        app_args=[CREATE_SIG, 1, 2, 0],
    )

    app_opt_in_tx = transaction.ApplicationOptInTxn(
        sender=sender, sp=suggested_params, index=farm_app_id
    )

    return [fund_tx, create_app_tx, app_opt_in_tx]


def fetch_escrow_by_id(
    algod: AlgodClient, app_id: int, farm: Optional["Farm"] = None
) -> "Escrow":
    state, creator = fetch_escrow_global_state(algod, app_id)
    if farm is None:
        from .farm import fetch_farm_by_id

        farm = fetch_farm_by_id(algod, state.master_app)

    assert (
        farm.app_id == state.master_app
    ), f'Escrow "{app_id}" doesn\'t match farm "{farm.app_id}".'

    return Escrow(
        algod=algod, app_id=app_id, user_address=creator, farm=farm, state=state
    )


def fetch_escrow_global_state(
    algod: AlgodClient, app_id: int
) -> tuple[EscrowInternalState, str]:
    app_info = algod.application_info(app_id)
    internal_state = parse_global_escrow_state(app_info["params"]["global-state"])
    creator = app_info["params"]["creator"]
    return internal_state, creator


def parse_global_escrow_state(raw_state: list) -> EscrowInternalState:
    state = parse_app_state(raw_state)
    return EscrowInternalState(master_app=state["MasterAppID"])


@dataclass
class Escrow:
    algod: AlgodClient
    app_id: int
    farm: "Farm"
    user_address: str
    state: EscrowInternalState
    _suggested_params: Optional[algosdk.transaction.SuggestedParams] = None
    address: str = field(init=False)

    def __post_init__(self):
        self.address = algosdk.logic.get_application_address(self.app_id)

    def __hash__(self):
        return self.app_id

    def __str__(self):
        return f"Escrow {self.user_address[:8]}... ({self.app_id}) for {self.farm}"

    def __repr__(self):
        return str(self)

    def __eq__(self, other_obj: object) -> bool:
        if not isinstance(other_obj, Escrow):
            return False
        return self.app_id == other_obj.app_id

    @property
    def suggested_params(self) -> algosdk.transaction.SuggestedParams:
        assert self._suggested_params is not None
        return self._suggested_params

    def set_suggested_params(
        self, suggested_params: algosdk.transaction.SuggestedParams
    ):
        self._suggested_params = suggested_params

    def refresh_suggested_params(self):
        self.set_suggested_params(self.algod.suggested_params())

    def fetch_user_state(self):
        return self.farm.fetch_user_state(self.user_address)

    def get_user_state_from_account_info(self, account_info: dict):
        return self.farm.get_user_state_from_account_info(account_info)

    def build_stake_txs(self, amount: int) -> list[transaction.Transaction]:
        transfer_tx = self.farm.staked_asset.build_transfer_tx(
            sender=self.user_address,
            receiver=self.address,
            amount=amount,
            suggested_params=self.suggested_params,
        )
        update_txs = self.farm.build_update_with_opcode_increase_txs(self)

        return [transfer_tx, *update_txs]

    def build_unstake_txs(self, amount: int) -> list[transaction.Transaction]:
        unstake_tx = transaction.ApplicationNoOpTxn(
            sender=self.user_address,
            index=self.app_id,
            foreign_apps=[self.farm.app_id],
            foreign_assets=[self.farm.staked_asset.index],
            app_args=[
                UNSTAKE_SIG,
                abi.UintType(8).encode(0),
                abi.UintType(64).encode(amount),
                abi.UintType(8).encode(1),
            ],
            sp=sp_fee(self.suggested_params, 3000),
        )

        txs: list[transaction.Transaction] = [unstake_tx]

        if increase_opcode_quota_tx := self.farm.build_update_increase_opcode_quota_tx(
            self.user_address
        ):
            txs.insert(0, increase_opcode_quota_tx)

        return txs

    def build_claim_rewards_tx(self) -> transaction.Transaction:
        return self.farm.build_claim_rewards_tx(self)

    def build_send_message_tx(
        self, address: str, message: str
    ) -> transaction.Transaction:
        encoded_message = message.encode()
        note = algosdk.abi.UintType(16).encode(len(encoded_message)) + encoded_message
        return transaction.ApplicationNoOpTxn(
            sender=self.user_address,
            index=self.app_id,
            app_args=[SEND_MESSAGE_SIG, 1, note],
            accounts=[address],
            sp=sp_fee(self.suggested_params, 2000),
        )

    def build_withdraw_algos(self) -> transaction.Transaction:
        return transaction.ApplicationNoOpTxn(
            sender=self.user_address,
            index=self.app_id,
            app_args=[WITHDRAW_ALGOS_SIG],
            sp=sp_fee(self.suggested_params, 2000),
        )

    def build_force_exit_tx(self) -> transaction.Transaction:
        return transaction.ApplicationClearStateTxn(
            sender=self.user_address,
            index=self.farm.app_id,
            sp=self.suggested_params,
        )

    def build_exit_tx(self) -> transaction.Transaction:
        return transaction.ApplicationCloseOutTxn(
            sender=self.user_address,
            index=self.farm.app_id,
            sp=self.suggested_params,
        )

    def build_delete_tx(self) -> transaction.Transaction:
        return transaction.ApplicationDeleteTxn(
            sender=self.user_address,
            index=self.app_id,
            foreign_apps=[self.farm.app_id],
            foreign_assets=[self.farm.staked_asset.index],
            sp=sp_fee(self.suggested_params, 3000),
        )



import base64
import dataclasses
from typing import Callable, Optional

import algosdk

from pactsdk.encoding import deserialize_uint64

from ..pool import Pool, fetch_pool_by_id
from ..transaction_group import TransactionGroup
from ..utils import get_box_min_balance, parse_app_state, wait_for_confirmation

Signer = Callable[[TransactionGroup], list[algosdk.transaction.SignedTransaction]]


def get_contract_deploy_cost(
    extra_pages: int, num_byte_slices: int, num_uint: int, is_algo: bool
):
    cost = 100_000 + 100_000 * extra_pages
    cost += num_byte_slices * 50_000
    cost += num_uint * 28_500
    cost += get_box_min_balance(32, 8)

    # exchange opt-ins & min balance
    cost += 300_000 if is_algo else 400_000

    return cost + 300000


@dataclasses.dataclass
class PoolParams:
    primary_asset_id: int
    secondary_asset_id: int
    fee_bps: int
    version: int

    abi = algosdk.abi.ArrayStaticType(algosdk.abi.UintType(64), 4)

    def __hash__(self):
        return hash(self.as_tuple())

    def as_tuple(self):
        return (
            self.primary_asset_id,
            self.secondary_asset_id,
            self.fee_bps,
            self.version,
        )

    def to_box_name(self) -> bytes:
        return self.abi.encode(self.as_tuple())

    @classmethod
    def from_box_name(cls, name: str):
        values = cls.abi.decode(base64.b64decode(name))
        return cls(*values)


@dataclasses.dataclass
class PoolBuildParams:
    primary_asset_id: int
    secondary_asset_id: int
    fee_bps: int


@dataclasses.dataclass
class FactoryState:
    pool_version: int
    allowed_fee_bps: list[int]


def parse_global_factory_state(raw_state: list) -> FactoryState:
    """
    Args:
        raw_state: The contract's global state retrieved from algosdk.
    """
    state = parse_app_state(raw_state)
    return FactoryState(
        pool_version=state["POOL_CONTRACT_VERSION"],
        allowed_fee_bps=deserialize_uint64(state["ALLOWED_FEE_BPS"]),
    )


def list_pools(
    algod: algosdk.v2client.algod.AlgodClient,
    factory_id: int,
) -> list[PoolParams]:
    boxes = algod.application_boxes(factory_id)
    return [PoolParams.from_box_name(box["name"]) for box in boxes["boxes"]]


def fetch_pool_id(
    algod: algosdk.v2client.algod.AlgodClient,
    factory_id: int,
    pool_params: PoolParams,
) -> int:
    box_name = pool_params.to_box_name()
    try:
        box = algod.application_box_by_name(factory_id, box_name)
    except algosdk.error.AlgodHTTPError as e:
        if "box not found" in str(e):
            return 0
        raise e
    return int.from_bytes(base64.b64decode(box["value"]), "big")


@dataclasses.dataclass
class PoolFactory:
    """Abstract class for pool factories.

    The pool factory allows decentralization of pools creation and discoverability.
    Each pool type has a separate factory contract that deploys the pool. Every pool created by the pool factory can be trusted as a valid Pact pool.

    The factory ensures pools uniqueness meaning you can't create two pools with the same parameters using a single factory contract.
    """

    algod: algosdk.v2client.algod.AlgodClient
    app_id: int
    state: FactoryState

    def list_pools(self) -> list[PoolParams]:
        """Lists all pools created by this factory. It works by reading the boxes created by this factory. The boxes serve as a hash map of unlimited size. The box name stores pool parameters and the box content stores pool id.

        This method returns only pool parameters without the application id. You have to call `fetch_pool` to fetch the actual pool e.g.

        pool_params = factory.list_pools()
        pool = factory.fetch_pool(pool_params[0])

        Returns:
            List of pool parameters.
        """
        return list_pools(self.algod, self.app_id)

    def fetch_pool_id(self, pool_params: PoolParams) -> int:
        return fetch_pool_id(self.algod, self.app_id, pool_params)

    def fetch_pool(self, pool_params: PoolParams) -> Optional[Pool]:
        """Fetches the pool for the given params.

        Args:
            pool_params: Parameters of the pool with are looking for.

        Returns:
            A pool if pool with given parameters exists, None otherwise.
        """
        pool_id = self.fetch_pool_id(pool_params)
        if pool_id == 0:
            return None
        return fetch_pool_by_id(self.algod, pool_id)

    def build(
        self, sender: str, pool_build_params: PoolBuildParams, signer: Signer
    ) -> Pool:
        """Deploys a new pool to the network.

        Args:
            sender: The address that is going to send the transactions.
            pool_build_params: Parameters of the pool that is going to be created.
            signer: A callback that allows signing the transaction.

        Returns:
            The created pool instance.
        """
        assert (
            pool_build_params.fee_bps in self.state.allowed_fee_bps
        ), f"only one of {self.state.allowed_fee_bps} is allowed for fee_bps"

        sp = self.algod.suggested_params()
        tx_group = self.build_tx_group(
            sender=sender,
            pool_build_params=pool_build_params,
            sp=sp,
        )
        signed_txs = signer(tx_group)
        self.algod.send_transactions(signed_txs)
        txid = tx_group.transactions[-1].get_txid()
        wait_for_confirmation(self.algod, txid)
        txinfo = self.algod.pending_transaction_info(txid)
        pool_id = txinfo["inner-txns"][0]["application-index"]
        return fetch_pool_by_id(self.algod, pool_id)

    def build_or_get(
        self, sender: str, pool_build_params: PoolBuildParams, signer: Signer
    ) -> tuple[Pool, bool]:
        """Deploys a new pool to the network if the pool with the specified params does not exist yet. Otherwise, it returns the existing pool.

        Args:
            sender: The address that is going to send the transactions.
            pool_build_params: Parameters of the pool that is going to be created.
            signer: A callback that allows signing the transaction.

        Returns:
            The two items tuple. The first item is the created or existing pool. The second item is True if a new pool is created or False if an existing pool is returned.
        """

        try:
            new_pool = self.build(sender, pool_build_params, signer)
            return new_pool, True
        except algosdk.error.AlgodHTTPError as e:
            existing_pool = self.fetch_pool(
                PoolParams(
                    **dataclasses.asdict(pool_build_params),
                    version=self.state.pool_version,
                )
            )
            if existing_pool:
                return existing_pool, False
            raise e

    def build_tx_group(
        self,
        sender: str,
        sp: algosdk.transaction.SuggestedParams,
        pool_build_params: PoolBuildParams,
    ):
        raise NotImplementedError



import algosdk

from ..config import Config
from ..pool import PoolType
from .base_factory import PoolFactory, parse_global_factory_state
from .constant_product import ConstantProductFactory


def get_pool_factory(
    algod: algosdk.v2client.algod.AlgodClient, pool_type: PoolType, config: Config
) -> PoolFactory:
    if pool_type == "CONSTANT_PRODUCT":
        app_id = config.factory_constant_product_id
        factory_class = ConstantProductFactory
    elif pool_type == "NFT_CONSTANT_PRODUCT":
        app_id = config.factory_nft_constant_product_id
        factory_class = ConstantProductFactory
    else:
        raise NotImplementedError(f"Factory for {pool_type} is not implemented yet.")

    assert app_id, f"Missing factory id for {pool_type} factory"
    app_info = algod.application_info(app_id)
    factory_state = parse_global_factory_state(app_info["params"]["global-state"])
    return factory_class(
        algod=algod,
        app_id=app_id,
        state=factory_state,
    )



import math
from typing import TYPE_CHECKING, Protocol

from pactsdk.constant_product_calculator import ConstantProductCalculator
from pactsdk.exceptions import PactSdkError
from pactsdk.stableswap_calculator import StableswapCalculator

from .asset import Asset

if TYPE_CHECKING:
    from .pool import Pool


class SwapCalculator(Protocol):
    pool: "Pool"

    def get_price(self, liq_a: float, liq_b: float) -> float:
        """Calculates the price of assets. Accepts and returns decimal values.

        Args:
            liq_a: Primary liquidity if calculating price for primary asset, secondary otherwise.
            liq_b: Secondary liquidity if calculating price for primary asset, primary otherwise.

        Returns:
            The price of one asset in relation to the other.
        """
        ...

    def get_swap_gross_amount_received(
        self,
        liq_a: int,
        liq_b: int,
        amount_deposited: int,
    ) -> int:
        """Converts amountDeposited to amountReceived. Ignores fee calculations.

        Args:
            liq_a: Primary liquidity if swapping primary asset, secondary otherwise.
            liq_b: Secondary liquidity if swapping primary asset, primary otherwise.
            amount_deposited: Amount of the asset deposited in the contract.

        Returns:
            Amount of asset received from the contract after swap.
        """
        ...

    def get_swap_amount_deposited(
        self,
        liq_a: int,
        liq_b: int,
        amount_received: int,
    ) -> int:
        """Converts amountReceived to amountDeposited. Ignores fee calculations.

        Args:
            liq_a: Primary liquidity if swapping primary asset, secondary otherwise.
            liq_b: Secondary liquidity if swapping primary asset, primary otherwise.
            amount_received: Amount of asset the user want to receive from the swap.

        Returns:
            Amount of the asset the user has to deposit in the contract.
        """
        ...

    def get_minted_liquidity_tokens(self, added_liq_a: int, added_liq_b: int) -> int:
        """Returns amount of liquidity tokens that are going to be minted when adding liquidity.

        Args:
            added_liq_a: Amount of primary asset to add to the pool.
            added_liq_b: Amount of secondary asset to add to the pool.

        Returns:
            Amount of liquidity tokens that will be minted and given to user.
        """
        ...


class PoolCalculator:
    """Contains functions for calculation statistics and other numerical data about the pool.

    The pool calculator uses internal data from the pool to calculate values like the Prices, Net Amounts and values for the swap. Uses different formulas based on pool type.
    """

    def __init__(self, pool: "Pool"):
        self.pool = pool

        self.swap_calculator: SwapCalculator
        if pool.pool_type in ["CONSTANT_PRODUCT", "NFT_CONSTANT_PRODUCT"]:
            self.swap_calculator = ConstantProductCalculator(pool)
        elif pool.pool_type == "STABLESWAP":
            self.swap_calculator = StableswapCalculator(pool)
        else:
            raise PactSdkError(f"Unknown pool type: ${pool.pool_type}")

    @property
    def primary_asset_amount(self) -> int:
        return self.pool.internal_state.A

    @property
    def secondary_asset_amount(self) -> int:
        return self.pool.internal_state.B

    @property
    def primary_asset_amount_decimal(self) -> float:
        return self.pool.internal_state.A / self.pool.primary_asset.ratio

    @property
    def secondary_asset_amount_decimal(self) -> float:
        return self.pool.internal_state.B / self.pool.secondary_asset.ratio

    @property
    def is_empty(self) -> bool:
        """Checks if the pool is currently empty.

        A pool is empty if either the primary or secondary asset is zero.

        Returns:
            True if the pool is empty, False otherwise.
        """
        return self.primary_asset_amount == 0 or self.secondary_asset_amount == 0

    @property
    def primary_asset_price(self) -> float:
        """
        Returns:
            Amount of secondary assets for a single primary asset.
        """
        return self.swap_calculator.get_price(
            self.primary_asset_amount_decimal, self.secondary_asset_amount_decimal
        )

    @property
    def secondary_asset_price(self) -> float:
        """
        Returns:
            Amount of primary assets for a single secondary asset.
        """
        return self.swap_calculator.get_price(
            self.secondary_asset_amount_decimal,
            self.primary_asset_amount_decimal,
        )

    def amount_deposited_to_net_amount_received(
        self, asset: Asset, amount_deposited: int
    ) -> int:
        """Converts amount deposited in the contract to amount received from the contract. Includes fee calculations.

        Args:
            asset: Asset to deposit in the contract.
            amount_deposited: Amount to deposit in the contract.

        Returns:
            The amount to receive from the contract.
        """
        gross_amount_received = self._amount_deposited_to_gross_amount_received(
            asset, amount_deposited
        )
        fee = self.get_fee_from_gross_amount(gross_amount_received)
        return gross_amount_received - fee

    def net_amount_received_to_amount_deposited(
        self, asset: Asset, net_amount_received: int
    ) -> int:
        """Converts amount received from the contract to amount deposited in the contract.

        Args:
            asset: Asset to deposit in the contract.
            amount_deposited: Amount to receive from the contract.

        Returns:
            The amount to deposit in the contract.
        """
        fee = self.get_fee_from_net_amount(net_amount_received)
        net_amount_received += fee
        return self._gross_amount_received_to_amount_deposited(
            asset, net_amount_received
        )

    def get_fee_from_gross_amount(self, gross_amount: int) -> int:
        """Calculates the fee from the gross amount based on pool's fee_bps.

        Args:
            gross_amount: The amount to receive from the contract not yet lessened by the fee.

        Returns:
            The calculated fee.
        """
        return gross_amount - (gross_amount * (10_000 - self.pool.fee_bps)) // 10_000

    def get_fee_from_net_amount(self, net_amount: int) -> int:
        """Calculates the fee from the net amount based on pool's fee_bps. This is used in the swap for exact calculations.

        Args:
            net_amount: The amount to receive from the contract already lessened by the fee.

        Returns:
            The calculated fee.
        """
        return math.ceil(
            net_amount / ((10_000 - self.pool.fee_bps) / 10_000) - net_amount
        )

    def _gross_amount_received_to_amount_deposited(
        self,
        asset: Asset,
        int_gross_amount_received: int,
    ) -> int:
        A, B = self.get_liquidities(asset)
        return self.swap_calculator.get_swap_amount_deposited(
            A,
            B,
            int_gross_amount_received,
        )

    def _amount_deposited_to_gross_amount_received(
        self,
        asset: Asset,
        amount_deposited: int,
    ) -> int:
        A, B = self.get_liquidities(asset)
        return self.swap_calculator.get_swap_gross_amount_received(
            A,
            B,
            amount_deposited,
        )

    def get_liquidities(self, asset: Asset) -> tuple[int, int]:
        """Returns the array of liquidities from the pool, sorting them by setting provided asset as primary.

        Args:
            asset: The asset that is supposed to be the primary one.

        Returns:
            Total liquidities of assets.
        """
        A, B = [self.primary_asset_amount, self.secondary_asset_amount]
        if asset != self.pool.primary_asset:
            A, B = B, A
        return A, B

    def get_minimum_amount_received(
        self, asset: Asset, amount: int, slippage_pct: float
    ) -> int:
        """Based on the deposited amount and a slippage, calculate the minimum amount the user will receive from the contract.

        Args:
            asset: The asset to deposit in the contract.
            amount_deposited: The amount to deposit in the contract.
            slippage_pct: Slippage in percents.

        Returns:
            The minimum amount to receive from the contract.
        """

        amount_received = self.amount_deposited_to_net_amount_received(asset, amount)
        return math.floor(amount_received - (amount_received * (slippage_pct / 100)))

    def get_fee(self, asset: Asset, amount_deposited: int) -> int:
        """Calculates the exchange fee based on deposited amount.

        Args:
            asset: The asset to deposit in the contract.
            amount_deposited: The amount to deposit in the contract.

        Returns:
            The calculated fee.
        """
        return self._amount_deposited_to_gross_amount_received(
            asset, amount_deposited
        ) - (self.amount_deposited_to_net_amount_received(asset, amount_deposited))

    def get_asset_price_after_liq_change(
        self,
        asset: Asset,
        primary_liq_change: int,
        secondary_liq_change: int,
    ) -> float:
        """Simulates new asset price after changing the pool's liquidity.

        Args:
            asset: The asset for which to calculate the price for.
            primary_liq_change: The change of primary liquidity on the pool.
            secondary_liq_change: The change of secondary liquidity on the pool.

        Returns:
            New asset price.
        """
        new_primary_liq = (
            self.primary_asset_amount + primary_liq_change
        ) / self.pool.primary_asset.ratio
        new_secondary_liq = (
            self.secondary_asset_amount + secondary_liq_change
        ) / self.pool.secondary_asset.ratio

        if asset == self.pool.primary_asset:
            return self.swap_calculator.get_price(new_primary_liq, new_secondary_liq)
        return self.swap_calculator.get_price(new_secondary_liq, new_primary_liq)

    def get_price_impact_pct(
        self,
        asset: Asset,
        primary_liq_change: int,
        secondary_liq_change: int,
    ) -> float:
        """Calculates the price impact of changing the liquidity in a certain way.

        Args:
            asset: The asset for which to calculate the price impact for.
            primary_liq_change: The change of primary liquidity on the pool.
            secondary_liq_change: The change of secondary liquidity on the pool.

        Returns:
            The asset price impact.
        """
        old_price = (
            self.primary_asset_price
            if asset == self.pool.primary_asset
            else self.secondary_asset_price
        )
        new_price = self.get_asset_price_after_liq_change(
            asset,
            primary_liq_change,
            secondary_liq_change,
        )
        return new_price / old_price * 100 - 100

    def get_swap_price(self, asset_deposited: Asset, amount_deposited: int) -> float:
        """Calculates the price for which the asset in going to be swapped.

        Args:
            asset_deposited: The asset deposited in the contract.
            amount_deposited: The amount deposited in the contract.

        Returns:
            The price of deposited asset in relation to received asset.
        """
        asset_received = self.pool.get_other_asset(asset_deposited)
        amount_received = self._amount_deposited_to_gross_amount_received(
            asset_deposited, amount_deposited
        )
        diff_ratio = asset_deposited.ratio / asset_received.ratio
        return amount_received / amount_deposited * diff_ratio



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pactsdk', 'pactsdk.factories', 'pactsdk.farming']

package_data = \
{'': ['*']}

install_requires = \
['py-algorand-sdk>=2.0.0,<3.0.0', 'requests>=2.27.1,<3.0.0']

setup_kwargs = {
    'name': 'pactsdk',
    'version': '0.6.1',
    'description': 'Python SDK for Pact smart contracts',
    'long_description': '# Pact Python SDK\n\n**pactsdk** is a software development kit for interfacing to [Pact](https://pact.fi), a decentralized automated market maker on the Algorand protocol.\n\nThe full documentation for this module can be found here:\n\n[https://pactfi.github.io/pact-py-sdk/latest/](https://pactfi.github.io/pact-py-sdk/latest/)\n\nThe Python SDK provides a set of modules on top of the Algorand Python SDK for interacting with liquidity pools and making swaps.\nClients can use the Python SDK to enhance their trading experience with Pact.\n\nWhat is covered by the library:\n\n- Creating pools\n- Managing liquidity\n- Making swaps\n- Farming\n\nSigning and sending transactions is not covered by the library. The provided examples use algosdk directly to send the transactions.\n\n# Installation\n\n`pip install pactsdk`\n\n# Basic usage\n\n**CAUTION** - The library uses integers for asset amounts e.g. microalgos instead of algos so if you want to send 1 algo, you need to specify it as 1_000_000.\n\nCreate a Pact client.\n\n```py\nfrom algosdk.v2client.algod import AlgodClient\nimport pactsdk\n\nalgod = AlgodClient(token, url)\npact = pactsdk.PactClient(algod)\n```\n\nBy default, the client is configured to work with mainnet. You can easily change it by providing `network` argument. The `network` argument changes the default values in `pact.config` object. It contains things like API URL or global contract ids.\n\n```py\npact = pactsdk.PactClient(algod, network="testnet")\n```\n\nFetching pools by assets pair. It uses Pact API to retrieve the pool. Can return multiple pools with differing fee_bps.\n\n```py\nalgo = pact.fetch_asset(0)\nother_coin = pact.fetch_asset(8949213)\n\npools = pact.fetch_pools_by_assets(algo, other_coin) # The pool will be fetched regardless of assets order.\n```\n\nYou can fetch a pool by providing assets ids instead of Asset objects.\n\n```py\npools = pact.fetch_pools_by_assets(0, 8949213)\n```\n\nYou can also fetch a pool by providing app id. This way the pool is retrieved directly from the chain.\n\n```py\npool = pact.fetch_pool_by_id(456321)\n```\n\nBefore making the transactions you need to opt-in for the assets. There\'s no need to opt-in for algo.\n\n```py\nimport algosdk\n\nprivate_key = algosdk.mnemonic.to_private_key(\'<mnemonic>\')\naddress = algosdk.account.address_from_private_key(private_key)\n\ndef opt_in(asset):\n    is_opted_in = asset.is_opted_in(address)\n    if not is_opted_in:\n        opt_in_tx = asset.prepare_opt_in_tx(address)\n        signed_tx = opt_in_tx.sign(private_key)\n        algod.send_transaction(signed_tx)\n\nopt_in(pool.primary_asset)\nopt_in(pool.secondary_asset)\nopt_in(pool.liquidity_asset) # Needed if you want to manage the liquidity.\n```\n\nCheck the current pool state.\n\n```py\nprint(pool.state)\n# PoolState(\n#   total_liquidity=900000,\n#   total_primary=956659,\n#   total_secondary=849972,\n#   primary_asset_price=0.8884795940873393,\n#   secondary_asset_price=1.1255182523659604,\n# )\n```\n\nExplicit pool state update is necessary periodically and after each pool operation.\n\n```py\npool.update_state()\npool.state  # Now holds fresh values.\n```\n\nManaging the liquidity.\n\n```py\n# Add liquidity.\nliquidity_addition = pool.prepare_add_liquidity(\n  primary_asset_amount=100_000,\n  secondary_asset_amount=50_000,\n);\nadd_liq_tx_group = liquidity_addition.prepare_tx_group(address)\nsigned_add_liq_tx_group = add_liq_tx_group.sign(private_key)\nalgod.send_transactions(signed_add_liq_tx_group)\n\n# Remove liquidity.\nremove_liq_tx_group = pool.prepare_remove_liquidity_tx_group(\n  address=address,\n  amount=100_000,\n)\nsigned_remove_liq_tx_group = remove_liq_tx_group.sign(private_key)\nalgod.send_transactions(signed_remove_liq_tx_group)\n```\n\nMaking a swap.\n\n```py\nswap = pool.prepare_swap(\n  asset=algo,\n  amount=200_000,\n  slippage_pct=2,\n)\n\n# You can inspect swap effect before submitting the transaction.\nprint(swap.effect)\n# SwapEffect(\n#     amount_deposited=200000,\n#     amount_received=146529,\n#     minimum_amount_received=143598,\n#     fee=441,\n#     price=0.73485,\n#     primary_asset_price_after_swap=0.6081680080300244,\n#     secondary_asset_price_after_swap=1.6442824791774173,\n#     primary_asset_price_change_pct=-31.549580645715963,\n#     secondary_asset_price_change_pct=46.091142966447585,\n# )\n\n# Let\'s submit the swap.\nswap_tx_group = swap.prepare_tx_group(address)\nsigned_tx_group = swap_tx_group.sign_txn(private_key)\nalgod.send_transactions(signed_tx_group)\n```\n\n## Composability of transactions.\n\nThe SDK has two sets of methods for creating transactions:\n\n1. `prepare_..._tx_group` e.g. `pool.prepare_swap_tx_group`\n\n   Those methods are convenience methods which ask algod for suggested transaction parameters, build transactions and create a transaction group. You can\'t add you own transactions to the group using those methods.\n\n2. `build_..._txs` e.g. `pool.build_swap_txs`\n\n   Those methods return a list of transactions. You can extend that list with your own transactions and create a `TransactionGroup` manually from this list.\n\n# Development\n\n- `poetry install`\n\nDevelopment requires [Pact contracts V1](https://github.com/pactfi/algorand-testbed) to be checked out.\n\n- `git clone git@github.com:pactfi/algorand-testbed.git`\n- `cd algorand-testbed`\n- `poetry install`\n- `docker compose up -d`\n- `cd ..`\n\n## Building\n\n- `poetry build`\n\nYou can install the package locally with\n`pip install dist/pactsdk-<version>.whl`\n\nValidate the installation `python -c "import pactsdk; print(pactsdk.__version__)"`\n\n## Running tests\n\n- `poetry run pytest`\n',
    'author': 'Mateusz Tomczyk',
    'author_email': 'mateusz.tomczyk@ulam.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/pactfi/pact-py-sdk',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import copy
import math
from dataclasses import dataclass, field
from typing import Literal, Optional, Union

import algosdk
from algosdk import transaction
from algosdk.v2client.algod import AlgodClient

from pactsdk.api import ListPoolsParams, list_pools
from pactsdk.constant_product_calculator import ConstantProductParams
from pactsdk.pool_state import (
    AppInternalState,
    PoolState,
    get_pool_type_from_internal_state,
    parse_global_pool_state,
)
from pactsdk.stableswap_calculator import StableswapParams

from .add_liquidity import LiquidityAddition
from .asset import Asset, fetch_asset_by_index
from .exceptions import PactSdkError
from .pool_calculator import PoolCalculator
from .swap import Swap
from .transaction_group import TransactionGroup
from .zap import Zap

PoolType = Literal["CONSTANT_PRODUCT", "NFT_CONSTANT_PRODUCT", "STABLESWAP"]

OperationType = Literal["SWAP", "ADDLIQ", "REMLIQ"]
"""The basic three operation types in a PACT liquidity pool, namely Add Liquidity (ADDLIQ), Remove Liquidity (REMLIQ) and making a swap (SWAP)."""


def fetch_app_global_state(
    algod: AlgodClient,
    app_id: int,
) -> AppInternalState:
    """Fetches the global state of the of an application.

    Args:
        algod: The algo client to query the app in.
        app_id: The application id to fetch the state of.

    Returns:
        The global state of the application.
    """
    app_info = algod.application_info(app_id)
    return parse_global_pool_state(app_info["params"]["global-state"])


def fetch_pool_by_id(algod: AlgodClient, app_id: int):
    """Fetches the pool from the blockchain using the provided algod client.

    Args:
        algod: The algo client to use.
        app_id: The application id to fetch.

    Returns:
        Pool: The pool object for the application id passed in.
    """
    app_global_state = fetch_app_global_state(algod, app_id)

    primary_asset = fetch_asset_by_index(algod, app_global_state.ASSET_A)
    secondary_asset = fetch_asset_by_index(algod, app_global_state.ASSET_B)
    liquidity_asset = fetch_asset_by_index(algod, app_global_state.LTID)

    return Pool(
        algod=algod,
        app_id=app_id,
        primary_asset=primary_asset,
        secondary_asset=secondary_asset,
        liquidity_asset=liquidity_asset,
        internal_state=app_global_state,
    )


def fetch_pools_by_assets(
    algod: AlgodClient,
    asset_a: Union[Asset, int],
    asset_b: Union[Asset, int],
    pact_api_url: str,
) -> list["Pool"]:
    """Returns the list of pools for the assets passed in.

    There can be zero pools if there are no pools matching the assets, or multiple if there are multiple at different fees.
    The order of assets that you provide is irrelevant.

    Args:
        algod: The algo client to use.
        asset_a: One of the assets in the pool (asset id or asset instance).
        asset_b: The other asset in the pool (asset id or asset instance).
        pact_api_url: The API url to use.

    Returns:
        A list of pools matching the provided assets.
    """
    assets = [
        asset.index if isinstance(asset, Asset) else asset
        for asset in [asset_a, asset_b]
    ]

    # Make sure that the user didn't mess up assets order.
    # Primary asset always has lower index.
    primary_asset, secondary_asset = sorted(assets)

    assert pact_api_url, "Must provide pact_api_url."

    app_ids = get_app_ids_from_assets(pact_api_url, primary_asset, secondary_asset)

    return [fetch_pool_by_id(algod, app_id) for app_id in app_ids]


def get_app_ids_from_assets(
    pact_api_url: str,
    primary_asset_index: int,
    secondary_asset_index: int,
) -> list[int]:
    """Returns the application ids for any pools that match the primary and secondary asset.

    This function finds any pools using the `pact_api_url` passed in that match the asset ids passed in.

    Args:
        pact_api_url: The API url to use.
        primary_asset_index: The asset id for the primary asset of the pool.
        secondary_asset_index: The asset id for the secondary asset of the pool.

    Returns:
        List of asset ids.
    """
    params: ListPoolsParams = {
        "primary_asset__algoid": primary_asset_index,
        "secondary_asset__algoid": secondary_asset_index,
    }
    data = list_pools(pact_api_url, params)
    return [int(pool["appid"]) for pool in data["results"]]


@dataclass
class Pool:
    """Pool represents a liquidity pool in the PACT AMM.

    Typically, users don't have to instantiate this class manually. Use :py:meth:`pactsdk.client.PactClient.fetch_pool_by_id` or :py:meth:`pactsdk.client.PactClient.fetch_pools_by_assets` instead.

    The primary methods of the pool are to create the transaction groups to enable you to:

    - Add Liquidity,
    - Removing Liquidity,
    - Create a Swap on the Pool.
    """

    algod: AlgodClient
    """The Algorand client to use."""

    app_id: int
    """The application id for the pool."""

    primary_asset: Asset
    """The asset of the liquidity pool with the lower index."""

    secondary_asset: Asset
    """The asset of the liquidity pool with the higher index."""

    liquidity_asset: Asset
    """The asset for the liquidity pool token (LP token) that is given when liquidity is added, and burned when liquidity is withdrawn."""

    internal_state: AppInternalState
    """The global state on the blockchain for this pool."""

    fee_bps: int = 30
    """The fee in basis points for swaps trading on the pool."""

    pool_type: PoolType = field(init=False)
    """Different pool types use different formulas for making swaps."""

    version: int = field(init=False)
    """The version of the contract. May be 0 for some old pools which don't expose the version in the global state."""

    def __post_init__(self):
        self.params: Union[StableswapParams, ConstantProductParams]

        self.pool_type = get_pool_type_from_internal_state(self.internal_state)

        if self.pool_type in ["CONSTANT_PRODUCT", "NFT_CONSTANT_PRODUCT"]:
            self.params = ConstantProductParams(
                fee_bps=self.internal_state.FEE_BPS,
                pact_fee_bps=self.internal_state.PACT_FEE_BPS or 0,
            )
        elif self.pool_type == "STABLESWAP":
            self.params = StableswapParams(
                fee_bps=self.internal_state.FEE_BPS,
                pact_fee_bps=self.internal_state.PACT_FEE_BPS or 0,
                initial_a=self.internal_state.INITIAL_A or 0,
                initial_a_time=self.internal_state.INITIAL_A_TIME or 0,
                future_a=self.internal_state.FUTURE_A or 0,
                future_a_time=self.internal_state.FUTURE_A_TIME or 0,
                precision=self.internal_state.PRECISION or 1,
            )
        else:
            raise PactSdkError(f'Unknown pool type "{self.pool_type}".')

        self.fee_bps = self.internal_state.FEE_BPS
        self.version = self.internal_state.VERSION or 0

        self.calculator = PoolCalculator(self)
        self.state = self.parse_internal_state(self.internal_state)

    def __eq__(self, other_pool: object) -> bool:
        """Return equal by comparing the pools app_id value."""
        if not isinstance(other_pool, Pool):
            return False
        return self.app_id == other_pool.app_id

    def get_escrow_address(self) -> str:
        """Get the escrow address of the pool.

        Returns:
            The address corresponding to that pools's escrow account.
        """
        return algosdk.logic.get_application_address(self.app_id)

    def get_other_asset(self, asset: Asset) -> Asset:
        """Returns the "other" asset, i.e. primary if secondary is passed in and vice versa.

        Args:
            asset: The primary or secondary asset of the pool.

        Raises:
            PactSdkError: If the asset passed in is not the primary or secondary asset.

        Returns:
            The other asset, if the primary asset was passed in it will be the secondary asset and vice versa.
        """
        if asset == self.primary_asset:
            return self.secondary_asset

        if asset == self.secondary_asset:
            return self.primary_asset

        raise PactSdkError(f"Asset with index {asset.index} is not a pool asset.")

    def update_state(self) -> PoolState:
        """Updates the internal and pool state properties by re-reading the global state in the blockchain.

        Updating the pool state is recommended if there is a pause between the construction of the pool and the creation of the transactions on the pool. Calling this method ensures that the the pool state is not stale.

        Returns:
            The new pool state.
        """
        self.internal_state = fetch_app_global_state(self.algod, self.app_id)
        self.state = self.parse_internal_state(self.internal_state)
        return self.state

    def prepare_add_liquidity(
        self,
        primary_asset_amount: int,
        secondary_asset_amount: int,
    ) -> LiquidityAddition:
        """
        Creates a new LiquidityAddition instance.

        Args:
            options: Options for adding the liquidity.

        Returns:
            A new LiquidityAddition object.
        """
        return LiquidityAddition(
            pool=self,
            primary_asset_amount=primary_asset_amount,
            secondary_asset_amount=secondary_asset_amount,
        )

    def prepare_add_liquidity_tx_group(
        self,
        address: str,
        liquidity_addition: LiquidityAddition,
    ) -> TransactionGroup:
        """Prepares a :py:class:`pactsdk.transaction_group.TransactionGroup` for adding liquidity to the pool. See :py:meth:`pactsdk.pool.Pool.buildAddLiquidityTxs` for details.

        Args:
            address: Account address that will deposit the primary and secondary assets and receive the LP token.
            primary_asset_amount: The amount of primary asset to deposit.
            secondary_asset_amount: The amount of secondary asset to deposit.

        Returns:
            A transaction group that when executed will add liquidity to the pool.
        """
        suggested_params = self.algod.suggested_params()
        txs = self.build_add_liquidity_txs(
            address, liquidity_addition, suggested_params
        )
        return TransactionGroup(txs)

    def build_add_liquidity_txs(
        self,
        address: str,
        liquidity_addition: LiquidityAddition,
        suggested_params: transaction.SuggestedParams,
    ) -> list[transaction.Transaction]:
        """Builds the transactions to add liquidity for the primary asset and secondary asset of the pool.

        In typical circumstances 3 transactions are generated:

        - deposit of asset A
        - deposit of asset B
        - "ADDLIQ" application call to add liquidity with the above deposits

        For constant product pools only - if the pool is empty and the product of both assets is larger or equal 2**64 than an additional set of 3 transactions is built.

        The initial liquidity must satisfy the expression `sqrt(a * b) - 1000 > 0`.

        Args:
            address: Account address that will deposit the primary and secondary assets and receive the LP token.
            primary_asset_amount: The amount of primary asset to deposit.
            secondary_asset_amount: The amount of secondary asset to deposit.
            suggested_params: Algorand suggested parameters for transactions.
            note: An optional note that can be added to the application ADDLIQ transaction.

        Raises:
            AssertionError: If initial liquidity is too low.

        Returns:
            List of transactions to add the liquidity.
        """
        primary_asset_amount = liquidity_addition.primary_asset_amount
        secondary_asset_amount = liquidity_addition.secondary_asset_amount

        initial_liq_txs: list[transaction.Transaction] = []
        if self.calculator.is_empty:
            assert (
                math.isqrt(primary_asset_amount * secondary_asset_amount) - 1000 > 0
            ), "Initial liquidity must satisfy the expression `sqrt(a * b) - 1000 > 0`"

            if self.pool_type == "CONSTANT_PRODUCT":
                # Adding initial liquidity has a limitation that the product of 2 assets must be lower than 2**64. Let's check if we can fit below the limit.
                max_product = 2**64
                product = primary_asset_amount * secondary_asset_amount
                if product >= max_product:
                    # Need to split the liquidity into two chunks.
                    divisor = int((product / max_product) ** 0.5 + 1)
                    primary_small_amount = primary_asset_amount // divisor
                    secondary_small_amount = secondary_asset_amount // divisor

                    primary_asset_amount -= primary_small_amount
                    secondary_asset_amount -= secondary_small_amount

                    initial_liq_txs = self.build_raw_add_liquidity_txs(
                        address=address,
                        primary_asset_amount=primary_small_amount,
                        secondary_asset_amount=secondary_small_amount,
                        suggested_params=suggested_params,
                        fee=liquidity_addition.effect.tx_fee,
                        note=b"Pact initial add liquidity",
                    )

        txs = self.build_raw_add_liquidity_txs(
            address=address,
            primary_asset_amount=primary_asset_amount,
            secondary_asset_amount=secondary_asset_amount,
            suggested_params=suggested_params,
            fee=liquidity_addition.effect.tx_fee,
        )

        return [*initial_liq_txs, *txs]

    def build_raw_add_liquidity_txs(
        self,
        address: str,
        primary_asset_amount: int,
        secondary_asset_amount: int,
        suggested_params: transaction.SuggestedParams,
        fee: int,
        note=b"",
    ):
        tx1 = self._make_deposit_tx(
            address=address,
            asset=self.primary_asset,
            amount=primary_asset_amount,
            note=note or b"Pact add liquidity deposit",
            suggested_params=suggested_params,
        )
        tx2 = self._make_deposit_tx(
            address=address,
            asset=self.secondary_asset,
            amount=secondary_asset_amount,
            note=note or b"Pact add liquidity deposit",
            suggested_params=suggested_params,
        )
        tx3 = self._make_application_noop_tx(
            address=address,
            fee=fee,
            args=["ADDLIQ", 0],
            extraAsset=self.liquidity_asset,
            suggested_params=suggested_params,
            note=note,
        )

        return [tx1, tx2, tx3]

    def prepare_remove_liquidity_tx_group(
        self, address: str, amount: int
    ) -> TransactionGroup:
        """Prepares the transaction group for removing liquidity from the pool.

        Args:
            address: Account address that will deposit the LP token and receive the primary and secondary assets.
            amount: The amount of the LP token to return to the pool.

        Returns:
            Transaction group that when executed will remove liquidity from the pool.
        """
        suggested_params = self.algod.suggested_params()
        txs = self.build_remove_liquidity_txs(address, amount, suggested_params)
        return TransactionGroup(txs)

    def build_remove_liquidity_txs(
        self, address: str, amount: int, suggested_params: transaction.SuggestedParams
    ) -> list[transaction.Transaction]:
        """This creates two transactions in a group for the remove operation.

        - deposit of the liquidity asset
        - "REMLIQ" application call to remove the LP token from the account and receive the deposited assets in return.

        Args:
            address: Account address that will deposit the LP token and receive the primary and secondary assets.
            amount: The amount of the LP token to return to the pool.
            suggested_params: Algorand suggested parameters for transactions.

        Returns:
            List of transactions to remove the liquidity.
        """
        tx1 = self._make_deposit_tx(
            address=address,
            amount=amount,
            asset=self.liquidity_asset,
            note=b"Pact remove liquidity deposit",
            suggested_params=suggested_params,
        )
        tx2 = self._make_application_noop_tx(
            address=address,
            fee=3000,
            args=["REMLIQ", 0, 0],  # min expected primary, min expected secondary
            suggested_params=suggested_params,
        )

        return [tx1, tx2]

    def prepare_swap(
        self, asset: Asset, amount: int, slippage_pct: float, swap_for_exact=False
    ) -> Swap:
        """Creates a new swap instance for receiving the amount of asset within the slippage percent from the pool.

        Args:
            asset: The asset to swap.
            amount: Amount to swap or to receive. Look at `swap_for_exact` flag for details.
            slippage_pct: The maximum allowed slippage in percents e.g. `10` is 10%. The swap will fail if the slippage will be higher.
            swap_for_exact: If false or not provided, the `amount` is the amount to swap (deposit in the contract). If true, the `amount` is the amount to receive from the swap.

        Returns:
            A new swap object.
        """
        assert self.is_asset_in_the_pool(asset), f"Asset {asset.index} not in the pool"
        return Swap(
            self,
            asset_deposited=asset,
            amount=amount,
            slippage_pct=slippage_pct,
            swap_for_exact=swap_for_exact,
        )

    def prepare_swap_tx_group(self, swap: Swap, address: str) -> TransactionGroup:
        """Prepares a transaction group that when executed will perform a swap on the pool.

        Args:
            swap: The swap for which to generate transactions.
            address: The address that is performing the swap.

        Returns:
            Transaction group that when executed will perform a swap on the pool.
        """
        suggested_params = self.algod.suggested_params()
        txs = self.build_swap_txs(swap, address, suggested_params)
        return TransactionGroup(txs)

    def build_swap_txs(
        self, swap: Swap, address: str, suggested_params: transaction.SuggestedParams
    ) -> list[transaction.Transaction]:
        """Builds two transactions:

        - deposit of the asset to swap
        - "SWAP' application call that performs the swap to receive the other asset

        Args:
            swap: The swap for which to generate transactions.
            address: The address that is performing the swap.
            suggested_params: Algorand suggested parameters for transactions.

        Returns:
            List of transactions to perform the swap.
        """
        tx1 = self._make_deposit_tx(
            address=address,
            amount=swap.effect.amount_deposited,
            asset=swap.asset_deposited,
            note=b"Pact swap deposit",
            suggested_params=suggested_params,
        )
        tx2 = self._make_application_noop_tx(
            address=address,
            fee=swap.effect.tx_fee,
            args=["SWAP", swap.effect.minimum_amount_received],
            suggested_params=suggested_params,
        )

        return [tx1, tx2]

    def is_asset_in_the_pool(self, asset: Asset) -> bool:
        """Check if the asset is the primary or secondary asset of this pool.

        Args:
            asset: The asset to check is in the pool.

        Returns:
            True if the asset is in the pool or False otherwise.
        """
        return asset.index in {self.primary_asset.index, self.secondary_asset.index}

    def parse_internal_state(self, state: AppInternalState) -> PoolState:
        """Read the new pool state from the global state of the application.

        Args:
            state: Global state for the application.

        Returns:
            Parsed state.
        """
        return PoolState(
            total_liquidity=state.L,
            total_primary=state.A,
            total_secondary=state.B,
            primary_asset_price=self.calculator.primary_asset_price,
            secondary_asset_price=self.calculator.secondary_asset_price,
        )

    def prepare_zap(self, asset: Asset, amount: int, slippage_pct: float) -> Zap:
        """Creates a new zap instance for getting all required data for performing a zap.

        Args:
            asset: The asset to zap.
            amount: Amount used for the zap.
            slippage_pct: The maximum allowed slippage in percents e.g. `10` is 10%. The swap will fail if the slippage will be higher.

        Returns:
            A new zap object.
        """
        return Zap(
            self,
            asset=asset,
            amount=amount,
            slippage_pct=slippage_pct,
        )

    def prepare_zap_tx_group(self, zap: Zap, address: str) -> TransactionGroup:
        """Prepares a transaction group that when executed will perform a Zap on the pool.

        Args:
            zap: The zap for which to generate transactions.
            address: The address that is performing the Zap.

        Returns:
            Transaction group that when executed will perform a Zap on the pool.
        """
        suggested_params = self.algod.suggested_params()
        txs = self.build_zap_txs(zap, address, suggested_params)
        return TransactionGroup(txs)

    def build_zap_txs(
        self, zap: Zap, address: str, suggested_params: transaction.SuggestedParams
    ) -> list[transaction.Transaction]:
        """Builds the transactions to perform a Zap on the pool as per the options passed in. Zap allows to add liquidity to the pool by providing only one asset.

        This function will generate swap Txs to get a proper amount of the second asset and then generate add liquidity Txs with both of those assets.
        See :py:meth:`pactsdk.pool.Pool.buildSwapTxs` and :py:meth:`pactsdk.pool.Pool.buildAddLiquidityTxs` for more details.

        This feature is supposed to work with constant product pools only. Stableswaps can accept one asset to add liquidity by default.

        Args:
            zap: The zap for which to generate transactions.
            address: The address that is performing the Zap.
            suggested_params: Algorand suggested parameters for transactions.

        Returns:
            List of transactions to perform the Zap.
        """
        swap_txs = self.build_swap_txs(zap.swap, address, suggested_params)
        add_liq_txs = self.build_add_liquidity_txs(
            address, zap.liquidity_addition, suggested_params
        )
        return [*swap_txs, *add_liq_txs]

    def _make_deposit_tx(
        self,
        asset: Asset,
        address: str,
        amount: int,
        note: bytes,
        suggested_params: transaction.SuggestedParams,
    ):
        return asset.build_transfer_tx(
            sender=address,
            receiver=self.get_escrow_address(),
            amount=amount,
            note=note,
            suggested_params=suggested_params,
        )

    def _make_application_noop_tx(
        self,
        address: str,
        args: list,
        fee: int,
        suggested_params: transaction.SuggestedParams,
        extraAsset: Optional[Asset] = None,
        note=b"",
    ):
        foreign_assets: list[int] = [
            self.primary_asset.index,
            self.secondary_asset.index,
        ]
        if extraAsset:
            foreign_assets.append(extraAsset.index)

        suggested_params = copy.copy(suggested_params)
        suggested_params.fee = fee
        suggested_params.flat_fee = True

        return transaction.ApplicationNoOpTxn(
            sender=address,
            index=self.app_id,
            foreign_assets=foreign_assets,
            app_args=args,
            sp=suggested_params,
            note=note,
        )



from dataclasses import dataclass, field
from typing import Optional

import algosdk
from algosdk import transaction

from .utils import get_selector, sp_fee

INCREASE_OPCODE_QUOTA_SIG = get_selector("increase_opcode_quota(uint64,uint64)void")


@dataclass
class GasStation:
    app_id: int
    app_address: str = field(init=False)

    def __post_init__(self):
        self.app_address = algosdk.logic.get_application_address(self.app_id)

    def build_fund_tx(
        self, sender: str, amount: int, suggested_params: transaction.SuggestedParams
    ) -> transaction.Transaction:
        return transaction.PaymentTxn(
            sender=sender,
            receiver=self.app_address,
            amt=amount,
            sp=suggested_params,
        )

    def build_increase_opcode_quota_tx(
        self,
        sender: str,
        count: int,
        suggested_params: transaction.SuggestedParams,
        extra_fee=0,
    ) -> transaction.Transaction:
        return transaction.ApplicationNoOpTxn(
            sender=sender,
            index=self.app_id,
            app_args=[INCREASE_OPCODE_QUOTA_SIG, count, 0],
            sp=sp_fee(suggested_params, (count + 1) * 1000 + extra_fee),
        )


_gas_station: Optional[GasStation] = None


def set_gas_station(app_id: int):
    global _gas_station
    _gas_station = GasStation(app_id)


def get_gas_station() -> GasStation:
    assert _gas_station, "Gas station not set. Use set_gas_station."
    return _gas_station
